""" SCADA Actor """
import functools
import logging
import random
import time
from typing import Optional
from typing import no_type_check

import dotenv
import gridworks.algo_utils as algo_utils
import pendulum
import requests
from algosdk.atomic_transaction_composer import TransactionWithSigner
from algosdk.future.transaction import *
from algosdk.v2client.algod import AlgodClient
from beaker.client import ApplicationClient
from gridworks.actor_base import ActorBase
from gridworks.algo_utils import BasicAccount
from gridworks.enums import GNodeRole
from gridworks.message import as_enum
from pydantic import BaseModel

import gwatn.api_types as api_types
import gwatn.config as config
from gwatn import DispatchContract
from gwatn.enums import AlgoCertType
from gwatn.enums import MessageCategorySymbol
from gwatn.enums import UniverseType
from gwatn.types import AtnParamsBrickstorageheater as AtnParams
from gwatn.types import DispatchContractConfirmed
from gwatn.types import DispatchContractConfirmed_Maker
from gwatn.types import GtDispatchBoolean
from gwatn.types import GtDispatchBoolean_Maker
from gwatn.types import GwCertId
from gwatn.types import GwCertId_Maker
from gwatn.types import HeartbeatA
from gwatn.types import HeartbeatA_Maker
from gwatn.types import HeartbeatB
from gwatn.types import HeartbeatB_Maker
from gwatn.types import JoinDispatchContract_Maker
from gwatn.types import ScadaCertTransfer_Maker
from gwatn.types import SimplesimDriverReport
from gwatn.types import SimplesimDriverReport_Maker
from gwatn.types import SimTimestep
from gwatn.types import SimTimestep_Maker
from gwatn.types import SnapshotBrickstorageheater as Snapshot


DISPATCH_CONTRACT_REPORTING_ALGOS = 5

LOG_FORMAT = (
    "%(levelname) -10s %(sasctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)

LOGGER.setLevel(logging.INFO)


def get_max_store_kwh_th(params: AtnParams) -> float:
    room_temp_c = (params.RoomTempF - 32) * 5 / 9
    return params.C * (params.MaxBrickTempC - room_temp_c)


class AtnHbStatus(BaseModel):
    LastHeartbeatReceivedMs: int
    AtnLastHex: Optional[str] = None
    ScadaLastHex: str = "0"


class SimpleScadaSimActorBase(ActorBase):
    def __init__(
        self,
        settings: config.ScadaSettings = config.ScadaSettings(
            _env_file=dotenv.find_dotenv()
        ),
    ):
        super().__init__(settings=settings)

        self.settings = settings
        self.atn_gni_id = settings.atn_gni_id
        self.acct: BasicAccount = BasicAccount(settings.sk.get_secret_value())
        self.client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )
        self.atn_params_type_name: str = "atn.params.brickstorageheater"
        self.sp = self.client.suggested_params()
        self.cert_id: Optional[GwCertId] = None
        self.talking_with: bool = False
        self.atn_hb_status = AtnHbStatus(
            LastHeartbeatReceivedMs=int(1000 * time.time())
        )
        # The dc_client (DispatchContract client) builds on top of the Algorand beaker ApplicationClient
        self.dc_client: Optional[ApplicationClient] = None
        self.dc_app_id: Optional[int] = None
        self.algo_init_check()
        self.universe_type = as_enum(
            self.settings.universe_type_value, UniverseType, UniverseType.default()
        )
        self._time: float = self.get_initial_time_s()
        self.api_type_maker_by_name = (
            api_types.TypeMakerByName
        )  # overwrites base class to include types used in this repo
        self.atn_params_type_name: str = "atn.params"  # overwrite in the subclass

    @property
    def atn_alias(self):
        """Removes `scada` from the end of the SCADA's GNodeAlias"""
        return self.alias[:-6]

    @property
    def ta_alias(self):
        """Removes `scada` from the end of the SCADA's GNodeAlias"""
        return self.atn_alias + ".ta"

    def local_rabbit_startup(self) -> None:
        rjb = MessageCategorySymbol.rjb.value
        tc_alias_lrh = self.settings.time_coordinator_alias.replace(".", "-")
        binding = f"{rjb}.{tc_alias_lrh}.timecoordinator.sim-timestep"

        cb = functools.partial(self.on_timecoordinator_bindok, binding=binding)
        self._consume_channel.queue_bind(
            self.queue_name, "timecoordinatormic_tx", routing_key=binding, callback=cb
        )
        LOGGER.info(
            f"Queue {self.queue_name} bound to timecoordinatormic_tx with {binding} "
        )

    @no_type_check
    def on_timecoordinator_bindok(self, _unused_frame, binding) -> None:
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")

    def time(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self._time
        else:
            return time.time()

    def get_initial_time_s(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self.settings.initial_time_unix_s
        else:
            return time.time()

    def prepare_for_death(self) -> None:
        self.actor_main_stopped = True

    ########################
    ## Receives
    ########################

    def route_message(
        self, from_alias: str, from_role: GNodeRole, payload: HeartbeatB
    ) -> None:
        if payload.TypeName == DispatchContractConfirmed_Maker.type_name:
            if from_role != GNodeRole.AtomicTNode:
                LOGGER.info(
                    f"Ignoring DispatchContractConfrimed from GNode with role {from_role}; expects AtomicTNode"
                )
            try:
                self.dispatch_contract_confirmed_received(payload)
            except:
                LOGGER.exception("Error in dispatch_contract_confirmed_received")
        elif payload.TypeName == GtDispatchBoolean_Maker.type_name:
            if from_role != GNodeRole.AtomicTNode:
                LOGGER.info(
                    f"Ignoring GtDispatchBooleanfrom GNode with role {from_role}; expects AtomicTNode"
                )
            try:
                self.dispatch_received(payload)
            except:
                LOGGER.exception("Error in dispatch_received")

        elif payload.TypeName == HeartbeatA_Maker.type_name:
            if from_role != GNodeRole.Supervisor:
                LOGGER.info(
                    f"Ignoring HeartbeatA from GNode with role {from_role}; expects Supervisor"
                )
            try:
                self.heartbeat_from_super(from_alias=from_alias, ping=payload)
            except:
                LOGGER.exception("Error in heartbeat_from_super")
        elif payload.TypeName == HeartbeatB_Maker.type_name:
            if from_role != GNodeRole.AtomicTNode:
                LOGGER.info(
                    f"Ignoring HeartbeatB from GNode with role {from_role}; expects AtomicTNode"
                )
            try:
                self.heartbeat_from_atn(ping=payload)
            except:
                LOGGER.exception("Error in heartbeat_from_atn")
        elif payload.TypeName == SimplesimDriverReport_Maker.type_name:
            try:
                self.simplesim_driver_report_received(payload)
            except:
                LOGGER.exception("Error in simplesim_driver_report_received")
        elif payload.TypeName == SimTimestep_Maker.type_name:
            if from_role != GNodeRole.TimeCoordinator:
                LOGGER.info(
                    f"Ignoring SimTimestep from GNode with role {from_role}; expects TimeCoordinator"
                )
            try:
                self.timestep_from_timecoordinator(payload)
            except:
                LOGGER.exception("Error in timestep_from_timecoordinator")

    def heartbeat_from_atn(self, ping: HeartbeatB) -> None:
        """
        This is the Scada's half of the DispatchContract Heartbeat pattern.
        It:
          - Checks that it has a DispatchContract (owns that SmartContract)
          - Checks the GNodeAlias and GNodeInstanceId to validate partner
          - Sends a reply HeartbeatB immediately back using a RabbitJsonDirect message
          - Sends an audit report of its action to the DispatchContract
            # TODO: save audit report for sending in a batch if SmartContract
            # exists but is not reachable (i.e. blockchain down)
        [more info](https://gridworks.readthedocs.io/en/latest/dispatch-contract.html)

        Args:
            payload (HeartbeatB): The latest heartbeat received from its
            AtomicTNode partner

        """
        self.ping = ping

        received_ms = int(time.time() * 1000)
        if not self.in_dispatch_contract():
            LOGGER.info(f"Not in Dispatch Contract. Ignoring")
        if ping.FromGNodeInstanceId != self.atn_gni_id:
            raise Exception("In dispatch contract but mismatched Atn Gni Id!")
        if self.atn_hb_status.AtnLastHex is not None:
            if ping.YourLastHex != self.atn_hb_status.ScadaLastHex:
                LOGGER.info("Received incorrect ping. Ignoring")
                return
        self.talking_with = True
        self.atn_hb_status.LastHeartbeatReceivedMs = int(time.time() * 1000)
        self.atn_hb_status.AtnLastHex = ping.MyHex
        self.atn_hb_status.ScadaLastHex = str(random.choice("0123456789abcdef"))
        LOGGER.info(f"Got {ping.MyHex}. Sending {self.atn_hb_status.ScadaLastHex}")
        pong = HeartbeatB_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            my_hex=self.atn_hb_status.ScadaLastHex,
            your_last_hex=self.atn_hb_status.AtnLastHex,
            last_received_time_unix_ms=self.atn_hb_status.LastHeartbeatReceivedMs,
            send_time_unix_ms=int(1000 * time.time()),
        ).tuple
        self.send_message(
            payload=pong, to_role=GNodeRole.AtomicTNode, to_g_node_alias=self.atn_alias
        )
        # Report to DispatchContract with heartbeat.algo.audit
        ptxn = PaymentTxn(self.acct.addr, self.sp, self.dc_client.app_addr, 1000)
        self.dc_client.call(
            DispatchContract.heartbeat_algo_audit,
            signed_proof=TransactionWithSigner(ptxn, self.acct.as_signer()),
            heartbeat=pong.as_dict(),
        )

    def heartbeat_from_super(self, from_alias: str, ping: HeartbeatA) -> None:
        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex=ping.MyHex
        ).tuple

        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )

        LOGGER.debug(
            f"[{self.alias}] Sent HB: SuHex {pong.YourLastHex}, AtnHex {pong.MyHex}"
        )

    def timestep_from_timecoordinator(self, payload: SimTimestep):
        if self._time == 0:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.info(f"TIME STARTED: {self.time_str()}")
        elif self._time < payload.TimeUnixS:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.debug(f"Time is now {self.time_str()}")
        elif self._time == payload.TimeUnixS:
            self.repeat_timestep(payload)

    ########################################################
    #  Related to Dispatch Contract and Scada Cert
    ########################################################

    def algo_init_check(self):
        """Upon startup, see if the dispatch contract has already been created
        and also make sure Scada is appropriately funded"""

        # First, request cert transfer. This is idempotent - if scada
        # addr already owns the ScadaCertificate it does nothing

        self.request_cert_transfer()
        apps = self.client.account_info(self.acct.addr)["created-apps"]
        balances = algo_utils.get_balances(self.client, self.acct.addr)
        micro_algos = balances[0]
        if micro_algos < 100_000:
            raise Exception(f"Insufficient funds to even opt into SCADA cert!")
        if len(apps) > 1:
            raise Exception("Should only be part of one app")
        if not self.dispatch_contract_created():
            # Create, fund and do bootstrap1 of the Dispatch contract
            dc_algos = DISPATCH_CONTRACT_REPORTING_ALGOS + (
                DispatchContract.min_balance / 1_000_000
            )
            if micro_algos < (dc_algos + 2) * 1_000_000:
                raise Exception(f"Insufficient funding! Need at least {dc_algos + 2}")
            self.dc_client = ApplicationClient(
                self.client, DispatchContract(), signer=self.acct.as_signer()
            )
            app_id, app_address, transaction_id = self.dc_client.create()
            self.dc_app_id = app_id
            LOGGER.info(f"Created Dispatch Contract, app id {self.dc_app_id}")
            sp = self.dc_client.get_suggested_params()
            sp.flat_fee = True
            sp.fee = 2000
            ptxn = PaymentTxn(
                self.acct.addr,
                sp,
                self.dc_client.app_addr,
                DispatchContract.min_balance,
            )

            # Now do the first half of bootstrapping the contract by providing it
            # with enough money to fund about 32kB in boxes, along with
            # our ScadaCertId (which can be shown to match our address in the sig)

            result = self.dc_client.call(
                DispatchContract.bootstrap1,
                scada_seed=TransactionWithSigner(ptxn, self.acct.as_signer()),
                ScadaCert=self.cert_id.Idx,
            )
            LOGGER.info("Called bootstrap1 method of DispatchContract")
            LOGGER.info(
                f"Providing the scada certificate, and funding with {dc_algos} Algos"
            )
            if result.return_value != self.ta_alias:
                raise Exception(
                    f"Expected {self.ta_alias} back from bootstrapping app {self.dc_app_id}, but"
                    f" got {result.return_value}"
                )
        else:
            created_apps = self.client.account_info(self.acct.addr)["created-apps"]
            self.dc_app_id = created_apps[0]["id"]
            self.dc_client = ApplicationClient(
                client=self.client,
                app=DispatchContract(),
                signer=self.acct.as_signer(),
                app_id=self.dc_app_id,
            )
            LOGGER.info(f"Dispatch contract already exists, app id {self.dc_app_id}")
        self.in_dispatch_contract()

    def request_cert_transfer(self) -> None:
        """
        Opts into certificate and then request transfer of the ScadaCert from the GNodeFactory

        Raises:
            Exception if the scada certificate does not exist, or transfer is not successful
        """
        cert_type = AlgoCertType(self.settings.cert_type_value)
        if cert_type == AlgoCertType.SmartSig:
            raise NotImplementedError("Not prepared for SmartSig Certificates yet")

        # Look for cert id
        gnf = self.settings.public.gnf_admin_addr
        a = self.client.account_info(gnf)["created-assets"]
        scada_nft_ids = list(
            filter(
                lambda x: x["params"]["unit-name"] == "SCADA"
                and x["params"]["name"] == self.ta_alias,
                a,
            )
        )
        if len(scada_nft_ids) == 0:
            raise Exception("Scada certificate does not exist")

        self.cert_id = GwCertId_Maker(
            type=AlgoCertType.ASA, idx=scada_nft_ids[0]["index"], addr=None
        ).tuple

        balances = algo_utils.get_balances(self.client, self.acct.addr)

        # First, opt into sig
        txn = AssetOptInTxn(self.acct.addr, self.sp, self.cert_id.Idx)

        signed_txn = txn.sign(self.acct.sk)
        if self.cert_id.Idx not in balances.keys():
            try:
                self.client.send_transaction(signed_txn)
            except:
                raise Exception(
                    "Failure sending transaction on Algo blockchain. Check sandbox and funding"
                )
            algo_utils.wait_for_transaction(self.client, signed_txn.get_txid())

        balances = algo_utils.get_balances(self.client, self.acct.addr)

        if balances[self.cert_id.Idx] == 1:
            # already own the cert
            return
        payload = ScadaCertTransfer_Maker(
            ta_alias="d1.isone.ver.keene.holly.ta",  # REPLACE WITH api_util.get_alias_from_scada_cert
            signed_proof=encoding.msgpack_encode(signed_txn),
        ).tuple

        api_endpoint = f"{self.settings.public.gnf_api_root}/scada-cert-transfer/"
        try:
            r = requests.post(url=api_endpoint, json=payload.as_dict())
        except:
            raise Exception(f"Post to {api_endpoint} failed")
        if r.status_code > 200:
            if r.status_code == 422:
                note = f"Error entering SLA: " + r.json()["detail"]
            else:
                note = r.reason
            raise Exception(f"Post to {api_endpoint} failed: {note} ")

        # check that scada now owns cert
        balances = algo_utils.get_balances(self.client, self.acct.addr)
        if balances[self.cert_id.Idx] != 1:
            raise Exception(f"Gnf claimed successful transfer but I do not own cert")

        LOGGER.info(f"Successfully received Scada Cert ASA {self.cert_id.Idx}")

    def initialize_dispatch_contract(self):
        """Starts up Dispatch contract to be used with AtomicTNode.


        Raises exception if Scada Addr
            - scada cert from settings matches what is in my acct
            - already has created an app
            - does not already have a unique Scada Cert ASA
        """

        if self.in_dispatch_contract():
            LOGGER.warning(
                f"Already in dispatch contract. Ignoring initialize_dispatch_contract!"
            )
            return

        balances = algo_utils.get_balances(self.client, self.acct.addr)
        non_algo_asas = list(set(balances.keys()) - {0})
        scada_certs = list(
            filter(
                lambda x: self.client.asset_info(x)["params"]["unit-name"] == "SCADA",
                non_algo_asas,
            )
        )

        if len(scada_certs) > 1:
            raise Exception(
                f"Scada {self.alias} has 2 scada certs. This should not happen!"
            )
        if len(scada_certs) == 0:
            raise Exception(
                "Scada acct needs to be created and funded by TaOwner, and Scada GNode "
                "needs to be authorized (with GNodeFactory sending Scada cert) "
                f"before starting up {self.alias}"
            )

        if self.cert_id.Idx != scada_certs[0]:
            raise Exception(
                f"Scada cert in settings {self.cert_id.Idx} does not match"
                f"scada cert in my acct {scada_certs[0]}"
            )

        txn = PaymentTxn(
            sender=self.acct.addr,
            receiver=self.dc_client.app_addr,
            sp=self.sp,
            amt=1000,
        )
        signed_txn = txn.sign(self.acct.sk)
        # Don't need to actually send it in order to establish signature
        payload = JoinDispatchContract_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            dispatch_contract_app_id=self.dc_app_id,
            signed_proof=encoding.msgpack_encode(signed_txn),
        ).tuple

        self.send_message(
            payload=payload,
            to_role=GNodeRole.AtomicTNode,
            to_g_node_alias=self.atn_alias,
        )
        LOGGER.info(f"Sent d = {payload.as_dict()}")

    def dispatch_contract_confirmed_received(self, payload: DispatchContractConfirmed):
        """Check the state of the app to confirm that the Atn has finished the bootstrap.
        Then
           - set the atn_gni_id to GNodeInstanceId from the payload (this is how the Scada
           will confirm identity of the AtomicTNode
           - set the atn_params to those shared by the AtomicTNode
           - opt into the Dispatch Contract (now the DispatchContract is live)"""
        if self.dc_app_id is None:
            raise Exception(
                "Thats odd ... dont have a Dispatch Contract App Id and I need to be"
                "the creator"
            )
        if self.in_dispatch_contract():
            LOGGER.warning(
                f"Ignoring DispatchContractConfirmed - already in dispatch contract {self.dc_app_id}"
            )
            return

        if payload.AtnParamsTypeName != self.atn_params_type_name:
            LOGGER.info(
                f"Unexpected AtnParamsTypeName {payload.AtnParamsTypeName}. Expects {self.atn_params_type_name}"
            )
            return
        self.atn_params = payload.Params
        app_state = self.dc_client.get_application_state()
        if "ta_trading_rights_idx" not in app_state.keys():
            LOGGER.warning(f"Atn bootstrap is not finished. Ignoring")
            return

        self.atn_gni_id = payload.FromGNodeInstanceId

        with open(".env", "a") as file:
            file.write(f'SCADA_ATN_GNI_ID="{payload.FromGNodeInstanceId}"\n')
        self.dc_client.opt_in()
        LOGGER.info(f"Dispatch Contract {self.dc_app_id} is live")

    #########################################################
    # Make the below into abstractmethods if pulling out base class
    #########################################################

    def simplesim_driver_report_received(self, payload: SimplesimDriverReport) -> None:
        """This gets received right before the top of the hour, from our
        best simulation of the TerminalAsset (which is happening in the
        AtomicTNode).

        Should be overwritten by the derived class, according to the
        DriverDataTypeName. Should then send snapshot."""
        if payload.FromGNodeInstanecId != self.atn_gni_id:
            LOGGER.info(f"Igoring {payload} - incorrect GNodeInstanceId")

    def send_snapshot(self):
        """Send a snapshot of current core sensed values to AtomicTNode.
        This is done every hour, and also on sensed power change."""
        if self.atn_params is None:
            return
        report_payload = Snapshot(
            FromGNodeAlias=self.alias,
            FromGNodeInstanceId=self.g_node_instance_id,
            PowerWatts=self.power_watts,
            StoreKwh=self.store_kwh,
            MaxStoreKwh=get_max_store_kwh_th(self.atn_params),
            AboutTerminalAssetAlias=self.ta_alias,
        )
        self.send_message(
            payload=report_payload,
            to_role=GNodeRole.AtomicTNode,
            to_g_node_alias=self.atn_alias,
        )

    def dispatch_received(self, payload: GtDispatchBoolean) -> None:
        """
        Dispatch received from AtomicTNode

          - Checks that the GNodeAlias and GNodeInstanceId belong to its
        AtomicTNode and that we have a dispatch contract
          - Sets talking_with to true
          - Follows instructions (turns on or turns off). Will turn on or
          off boost unless AboutNodeName is "a.element"

        For hourly sim:
          - Updatespower
          - Send status to AtomicTNode
        """
        if self.atn_params is None or self.in_dispatch_contract() is False:
            LOGGER.info("Igoring dispatch command, DispatchContract is not started")
        if payload.FromGNodeInstanceId != self.atn_gni_id:
            LOGGER.info(f"Igoring {payload}, not my Atn's GNodeInstanceId")
        self.talking_with = True
        if payload.AboutNodeName == "a.elements":
            # Making the grossly simplifying assumption that the heat pump turns on immediately
            if payload.RelayState == 1:
                new_power_watts = self.atn_params.RatedMaxPowerKw * 1000
            else:
                new_power_watts = 0
            if self.power_watts != new_power_watts:
                self.power_watts = new_power_watts
                self.send_snapshot()

    def new_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("New timestep")
        if not self.in_dispatch_contract():
            self.initialize_dispatch_contract()

    def repeat_timestep(self, payload: SimTimestep) -> None:
        LOGGER.info("Timestep received again")

    def time_str(self) -> str:
        return pendulum.from_timestamp(self.time()).strftime("%m/%d/%Y, %H:%M")

    def dispatch_contract_created(self) -> bool:
        """The SCADA only creates one kind of smart contract, which is its dispatch contract.
        So this just checks for a created app
        """
        created_apps = self.client.account_info(self.acct.addr)["created-apps"]
        if len(created_apps) > 1:
            raise NotImplementedError(
                f"SCADA not designed yet to create multiple dispatch contracts"
            )
        if len(created_apps) > 0:
            return True
        return False

    def in_dispatch_contract(self) -> bool:
        """Checks that bootstrap 2 was completed w addition of ta_trading_rights_idx,
        and also that Scada is opted in"""
        app_state = self.dc_client.get_application_state()
        if "ta_trading_rights_idx" not in app_state.keys():
            LOGGER.info(
                "Not in Dispatch Contract because ta_trading_rights_idx isn't in app_state"
            )
            LOGGER.info(
                "AtomicTNode has not finished its part in bootstrapping the contract"
            )
            return False
        # Todo: also check that atn is opted in
        apps = self.client.account_info(self.acct.addr)["apps-local-state"]
        app_ids = list(map(lambda x: x["id"], apps))
        if self.dc_client.app_id in app_ids:
            return True
        LOGGER.info(
            "Not in Dispatch Contract because scada has not yet opted into contract"
        )
        return False



import time

import gridworks.conversion_factors as cf
import pendulum
import xlsxwriter
from satn.strategies.heatpumpwithbooststore.flo import HeatPumpWithBoostStore__Flo
from satn.strategies.heatpumpwithbooststore.node import (
    Node_SpaceHeat__WaterStore as Node,
)

import gwatn.brick_storage_heater.strategy_utils as strategy_utils
from gwatn.brick_storage_heater.edge import Edge__BrickStorageHeater as Edge
from gwatn.enums import RecognizedCurrencyUnit
from gwatn.types import AtnParamsBrickstorageheater as AtnParams


OUTPUT_FOLDER = "output_data"

ON_PEAK_DIST_PRICE_PER_MWH_CUTOFF = 100
SHOULDER_PEAK_DIST_PRICE_PER_MWH_CUTOFF = 50


def export_xlsx(alias: str, flo: HeatPumpWithBoostStore__Flo, atn_params: AtnParams):
    local_start = pendulum.timezone(flo.timezone_string).convert(flo.flo_start_utc)
    date_str = local_start.strftime("%Y%m%d")
    hour_str = local_start.strftime("%H")
    file = (
        OUTPUT_FOLDER + f"/result_{alias}_{date_str}_{hour_str}_{int(time.time())}.xlsx"
    )
    # flo.graph_strategy_alias,

    file = file.lower()
    print(file)

    workbook = xlsxwriter.Workbook(file)
    starting_store_idx = flo.starting_store_idx

    # Add to gsr for more blank rows
    gsr = 30
    w = export_best_path_info(
        alias=alias, flo=flo, workbook=workbook, starting_store_idx=starting_store_idx
    )

    export_flo_graph(
        flo=flo,
        workbook=workbook,
        worksheet=w,
        starting_store_idx=starting_store_idx,
        graph_start_row=gsr,
    )

    export_params_xlsx(flo=flo, atn_params=atn_params, workbook=workbook)
    workbook.close()


def export_best_path_info(
    alias: str,
    flo: HeatPumpWithBoostStore__Flo,
    workbook: xlsxwriter.workbook.Workbook,
    starting_store_idx: int,
):
    w = workbook.add_worksheet(
        f"start {100 * starting_store_idx / flo.params.StorageSteps}%"
    )
    w.freeze_panes(0, 2)
    title_format = workbook.add_format({"bold": True})
    title_format.set_font_size(14)
    bold_format = workbook.add_format({"bold": True})
    gray_filler_format = workbook.add_format({"bg_color": "#edf0f2"})
    header_format = workbook.add_format({"bold": True, "text_wrap": True})
    mwh_format = workbook.add_format({"bold": True, "num_format": '0.00" MWh"'})
    if flo.params.CurrencyUnit == RecognizedCurrencyUnit.USD:
        currency_format = workbook.add_format({"num_format": "$#,##0.00"})
        currency_bold_format = workbook.add_format(
            {"bold": True, "num_format": "$#,##0.00"}
        )
    elif flo.params.CurrencyUnit == RecognizedCurrencyUnit.GBP:
        currency_format = workbook.add_format(
            {"num_format": "_-[$£-en-GB]* #,##0.00_-"}
        )
        currency_bold_format = workbook.add_format(
            {"bold": True, "num_format": "_-[$£-en-GB]* #,##0.00_-"}
        )

    w.set_column("A:A", 26)
    w.set_column("B:B", 15)
    OPT_PATH_STATE_VAR_ROW = 14

    swt_list = flo_utils.get_source_water_temp_f_list(flo.params)
    w.write(0, 0, f"GNode Alias: {alias}", title_format)
    w.write(1, 0, flo.graph_strategy_alias)
    w.write(2, 0, f"FLO start {flo.params.TimezoneString} ", header_format)
    local_start = pendulum.timezone(flo.timezone_string).convert(flo.flo_start_utc)
    w.write(2, 1, local_start.strftime("%Y/%m/%d %H:%M"))

    w.write(3, 0, "Total hours", header_format)
    w.write(3, 1, sum(flo.slice_duration_hrs), bold_format)

    w.write(4, 0, "Rt Energy Price ($/MWh)", header_format)
    w.write(
        4,
        1,
        sum(flo.RealtimeElectricityPrice) / len(flo.RealtimeElectricityPrice),
        currency_bold_format,
    )
    # w.write(5, 0, "Flat rate for hp ($/MWh)", header_format)
    # if flo.params.IsRegulating:
    #     w.write(5, 0, "Regulation Price ($/MWh)", header_format)
    #     w.write(
    #         5, 1, sum(flo.reg_price_per_mwh) / len(flo.reg_price_per_mwh), currency_bold_format
    #     )

    w.write(6, 0, "Dist Price ($/MWh)", header_format)
    w.write(
        6,
        1,
        sum(flo.DistributionPrice) / len(flo.DistributionPrice),
        currency_bold_format,
    )
    w.write(7, 0, "Outside Temp F", header_format)
    w.write(
        7,
        1,
        round(sum(flo.params.OutsideTempF) / len(flo.params.OutsideTempF), 2),
        bold_format,
    )
    w.write(8, 0, "COP", header_format)
    avg_cop = sum(flo.cop.values()) / len(flo.cop)
    w.write(8, 1, round(avg_cop, 2), bold_format)
    w.write(9, 0, "House Power Required AvgKw", header_format)
    w.write(
        9,
        1,
        round(sum(flo.params.PowerRequiredByHouseFromSystemAvgKwList), 2),
        bold_format,
    )
    w.write(10, 0, "Required Source Water Temp F", header_format)
    avg_swt = round((sum(swt_list) / len(swt_list)), 0)
    w.write(10, 1, avg_swt, bold_format)
    w.write(11, 0, "Max HeatPump kWh thermal", header_format)
    avg_max_thermal_hp_kwh = round(
        (sum(flo.max_thermal_hp_kwh.values()) / len(flo.max_thermal_hp_kwh)), 2
    )
    w.write(11, 1, avg_max_thermal_hp_kwh, bold_format)

    w.write(12, 0, "Outputs", header_format)

    for jj in range(flo.time_slices):
        hours_since_start = sum(flo.slice_duration_hrs[0:jj])
        local_time = local_start.add(hours=hours_since_start)
        w.write(2, jj + 2, local_time.strftime("%m/%d"))
        w.write(3, jj + 2, local_time.strftime("%H:%M"))
        w.write(4, jj + 2, flo.RealtimeElectricityPrice[jj], currency_format)
        if flo.params.IsRegulating:
            w.write(5, jj + 2, flo.reg_price_per_mwh[jj], currency_format)
        else:
            w.write(5, jj + 2, "", gray_filler_format)
        dp = flo.DistributionPrice[jj]
        LIGHT_GREEN_HEX = "#bbe3a6"
        LIGHT_RED_HEX = "#ff6363"
        if dp > ON_PEAK_DIST_PRICE_PER_MWH_CUTOFF:
            if flo.params.CurrencyUnit == RecognizedCurrencyUnit.USD:
                dist_format = workbook.add_format(
                    {"bg_color": LIGHT_RED_HEX, "num_format": "$#,##0.00"}
                )
            elif flo.params.CurrencyUnit == RecognizedCurrencyUnit.GBP:
                dist_format = workbook.add_format(
                    {
                        "bg_color": LIGHT_RED_HEX,
                        "num_format": "_-[$£-en-GB]* #,##0.00_-",
                    }
                )
        elif dp > SHOULDER_PEAK_DIST_PRICE_PER_MWH_CUTOFF:
            if flo.params.CurrencyUnit == RecognizedCurrencyUnit.USD:
                dist_format = workbook.add_format(
                    {"bg_color": "yellow", "num_format": "$#,##0.00"}
                )
            elif flo.params.CurrencyUnit == RecognizedCurrencyUnit.GBP:
                dist_format = workbook.add_format(
                    {"bg_color": "yellow", "num_format": "_-[$£-en-GB]* #,##0.00_-"}
                )
        else:
            if flo.params.CurrencyUnit == RecognizedCurrencyUnit.USD:
                dist_format = workbook.add_format(
                    {"bg_color": LIGHT_GREEN_HEX, "num_format": "$#,##0.00"}
                )
            elif flo.params.CurrencyUnit == RecognizedCurrencyUnit.GBP:
                dist_format = workbook.add_format(
                    {
                        "bg_color": LIGHT_GREEN_HEX,
                        "num_format": "_-[$£-en-GB]* #,##0.00_-",
                    }
                )
        w.write(6, jj + 2, flo.DistributionPrice[jj], dist_format)
        w.write(7, jj + 2, flo.params.OutsideTempF[jj])
        w.write(8, jj + 2, flo.cop[jj])
        w.write(
            9, jj + 2, round(flo.params.PowerRequiredByHouseFromSystemAvgKwList[jj], 2)
        )
        w.write(10, jj + 2, round(swt_list[jj], 0))
        w.write(11, jj + 2, round(flo.max_thermal_hp_kwh[jj], 2))

        w.write(12, jj + 2, "", gray_filler_format)

    node: Node = flo.node[0][starting_store_idx]
    w.write(OPT_PATH_STATE_VAR_ROW, 0, "Store Temp (F)", header_format)
    w.write(OPT_PATH_STATE_VAR_ROW + 1, 0, "HeatPump kWh thermal", header_format)
    w.write(OPT_PATH_STATE_VAR_ROW + 2, 0, "HeatPump kWh electric", header_format)
    w.write(OPT_PATH_STATE_VAR_ROW + 3, 0, "Boost kWh electric", header_format)
    w.write(OPT_PATH_STATE_VAR_ROW + 4, 0, "Energy cost (¢)", header_format)
    w.write(OPT_PATH_STATE_VAR_ROW + 5, 0, "Hours Since Start", header_format)

    store_temp_f = []
    opt_heatpump_electricity_used_kwh = []
    opt_boost_electricity_used_kwh = []
    opt_energy_cost_dollars = []

    best_idx = starting_store_idx
    dist_cost = []
    min_dist_price_per_mwh = min(flo.DistributionPrice)

    for jj in range(flo.time_slices):
        edge: Edge = flo.best_edge[node]
        store_temp_f.append(node.store_avg_water_temp_f)
        hp_kwh = edge.hp_electricity_avg_kw
        boost_kwh = edge.boost_electricity_used_avg_kw
        opt_heatpump_electricity_used_kwh.append(hp_kwh)
        opt_boost_electricity_used_kwh.append(boost_kwh)
        opt_energy_cost_dollars.append(edge.cost)
        hours_since_start = sum(flo.slice_duration_hrs[0:jj])

        w.write(OPT_PATH_STATE_VAR_ROW, jj + 2, round(node.store_avg_water_temp_f, 2))
        w.write(
            OPT_PATH_STATE_VAR_ROW + 1,
            jj + 2,
            round(edge.hp_thermal_energy_generated_avg_kw, 3),
        )
        w.write(
            OPT_PATH_STATE_VAR_ROW + 2, jj + 2, round(edge.hp_electricity_avg_kw, 3)
        )
        w.write(
            OPT_PATH_STATE_VAR_ROW + 3,
            jj + 2,
            round(edge.boost_electricity_used_avg_kw, 3),
        )
        w.write(OPT_PATH_STATE_VAR_ROW + 4, jj + 2, round(edge.cost * 100, 2))
        w.write(OPT_PATH_STATE_VAR_ROW + 5, jj + 2, round(hours_since_start, 1))
        node = flo.node[jj + 1][edge.end_idx]

    w.write(
        OPT_PATH_STATE_VAR_ROW,
        1,
        round(sum(store_temp_f) / len(store_temp_f), 0),
        bold_format,
    )
    w.write(
        OPT_PATH_STATE_VAR_ROW + 2,
        1,
        sum(opt_heatpump_electricity_used_kwh) / 1000,
        mwh_format,
    )
    w.write(
        OPT_PATH_STATE_VAR_ROW + 3,
        1,
        sum(opt_boost_electricity_used_kwh) / 1000,
        mwh_format,
    )
    w.write(
        OPT_PATH_STATE_VAR_ROW + 4,
        1,
        sum(opt_energy_cost_dollars),
        currency_bold_format,
    )

    w.write("H1", "Electricity cost of this path")
    total_cost = sum(opt_energy_cost_dollars)
    w.write("G1", total_cost, currency_bold_format)
    w.write("H2", "Total electricity MWh")
    total_electricity_mwh = (
        sum(opt_boost_electricity_used_kwh) + sum(opt_heatpump_electricity_used_kwh)
    ) / 1000
    w.write("G2", total_electricity_mwh, mwh_format)

    total_btu = sum(flo.params.PowerRequiredByHouseFromSystemAvgKwList) * cf.BTU_PER_KWH
    gallons_oil = total_btu / cf.BTU_PER_GALLON_OF_OIL / 0.85
    # assumes 85% efficient oil boiler

    w.write("L1", "Gallons of Oil")
    w.write("K1", round(gallons_oil), bold_format)

    w.write("L2", "Equivalent Price of Oil")
    w.write("K2", total_cost / gallons_oil, currency_bold_format)
    # w.write("H3", "Flat rate comparison")

    return w


def export_flo_graph(
    flo: HeatPumpWithBoostStore__Flo,
    workbook: xlsxwriter.workbook.Workbook,
    worksheet: xlsxwriter.workbook.Worksheet,
    starting_store_idx: int,
    graph_start_row: int,
):
    # worksheet.freeze_panes(graph_start_row - 1, 2)
    header_format = workbook.add_format({"bold": True, "text_wrap": True})
    percent_format = workbook.add_format({"num_format": '00.0"%"'})
    best_path_format = workbook.add_format({"bold": True, "bg_color": "#CDEBA6"})
    best_idx = starting_store_idx

    worksheet.write(graph_start_row - 1, 0, "Percent full", header_format)

    for mm in range(flo.params.StorageSteps + 1):
        kk = flo.params.StorageSteps - mm
        percent = round(100 * kk / flo.params.StorageSteps, 1)
        worksheet.write(graph_start_row + mm, 0, percent, percent_format)
    for jj in range(flo.time_slices):
        best_node = flo.node[jj][best_idx]
        for mm in range(flo.params.StorageSteps + 1):
            kk = flo.params.StorageSteps - mm
            node = flo.node[jj][kk]
            if kk == best_idx:
                try:
                    worksheet.write(
                        graph_start_row + mm,
                        jj + 2,
                        -round(node.path_benefit, 4),
                        best_path_format,
                    )
                except:
                    print(f"failed for best {jj},{kk}")
            else:
                try:
                    worksheet.write(
                        graph_start_row + mm, jj + 2, round(node.path_cost, 4)
                    )
                except:
                    print(f"failed for {jj},{kk}")
        best_idx = flo.best_edge[best_node].end_idx


def export_params_xlsx(
    flo: HeatPumpWithBoostStore__Flo,
    atn_params: AtnParams,
    workbook: xlsxwriter.workbook.Workbook,
):
    bold = workbook.add_format({"bold": True})
    w = workbook.add_worksheet("Params")
    derived_format_bold = workbook.add_format({"bold": True, "font_color": "green"})
    derived_format = workbook.add_format({"font_color": "green"})
    swt_list = flo_utils.get_source_water_temp_f_list(flo.params)
    w.set_column("A:A", 31)
    w.set_column("D:D", 31)
    w.set_column("G:G", 31)
    w.write("A1", "Key Parameters", bold)

    t = flo.params.OutsideTempF
    w.write("A4", "This Run ColdestTempF ", derived_format_bold)
    w.write("B4", min(t), derived_format)

    w.write("A5", "HouseWorstCaseTempF ", bold)
    w.write("B5", atn_params.HouseWorstCaseTempF)

    w.write("A7", "SystemMaxHeatOutputKwAvg", bold)
    w.write("B7", round(flo.params.SystemMaxHeatOutputKwAvg, 2))

    p = flo.params.PowerRequiredByHouseFromSystemAvgKwList
    w.write("A8", "This Run MaxHeatOutputKwAvg", derived_format_bold)
    w.write("B8", round(max(p), 2), derived_format)

    house_wc_kw = flo_utils.get_house_worst_case_heat_output_avg_kw(flo.params)
    w.write("A9", "HouseWorstCaseHeatOuputAvgKw", derived_format_bold)
    w.write("B9", round(house_wc_kw, 1), derived_format)

    w.write("A10", "HouseWorstCaseHeatOuput BTU/hr", derived_format_bold)
    w.write("B10", round(house_wc_kw * cf.BTU_PER_KWH), derived_format)

    w.write("A12", "EmitterMaxSafeSwtF", bold)
    w.write("B12", flo.params.EmitterMaxSafeSwtF)

    w.write("A13", "This Run SystemMaxHeatOutputSwtF", bold)
    w.write("B13", round(max(swt_list)))

    w.write("A14", "SystemMaxHeatOutputSWTF ", bold)
    w.write("B14", flo.params.SystemMaxHeatOutputSwtF)

    w.write("A15", "HeatPumpMaxWaterTempF ", bold)
    w.write("B15", flo.params.MaxHeatpumpSourceWaterTempF)

    w.write("A16", "RatedHeatpumpElectricityKw", bold)
    w.write("B16", flo.params.RatedHeatpumpElectricityKw)

    w.write("A17", "StoreMaxPowerKw", bold)
    w.write("B17", flo.params.StoreMaxPowerKw)

    if flo.params.EmitterPumpFeedbackModel == EmitterPumpFeedbackModel.ConstantDeltaT:
        w.write("A19", "SystemMaxHeatOutputDeltaTempF", bold)
        w.write("B19", flo.params.SystemMaxHeatOutputDeltaTempF)

        w.write("A20", "SystemMaxHeatOutputGpm", derived_format_bold)
        w.write("B20", round(flo.params.SystemMaxHeatOutputGpm, 2), derived_format)
    else:
        w.write("A19", "SystemMaxHeatOutputDeltaTempF", bold)
        w.write("B19", flo.params.SystemMaxHeatOutputDeltaTempF)

        w.write("A20", "SystemMaxHeatOutputGpm", derived_format_bold)
        w.write("B20", round(flo.params.SystemMaxHeatOutputGpm, 2), derived_format)

    w.write("A21", "Cop1TempF", bold)
    w.write("B21", flo.params.Cop1TempF)

    w.write("A22", "Cop4TempF", bold)
    w.write("B22", flo.params.Cop4TempF)

    w.write("A23", "StorePassiveLossRatio", bold)
    w.write("B23", flo.params.StorePassiveLossRatio)

    w.write("A25", "StorageSteps", bold)
    w.write("B25", flo.params.StorageSteps)

    #############

    annual_kwh = atn_params.AnnualHvacKwhTh
    annual_btu = round(cf.BTU_PER_KWH * annual_kwh)
    w.write("D3", "Annual HVAC kWhTh", bold)
    w.write("E3", annual_kwh)

    w.write("D4", "Annual HVAC MBTU", derived_format_bold)
    w.write("E4", round(annual_btu / 10**6), derived_format)

    w.write("D6", "StoreSizeGallons", bold)
    w.write("E6", flo.params.StoreSizeGallons)

    w.write("D7", "MaxStoreTempF", bold)
    w.write("E7", flo.params.MaxStoreTempF)

    w.write("D8", "ZeroPotentialEnergyWaterTempF", bold)
    w.write("E8", flo.params.ZeroPotentialEnergyWaterTempF)

    w.write("D9", "TotalStorageKwh", derived_format_bold)
    w.write("E9", round(flo.max_energy_kwh_th, 1), derived_format)

    w.write("D10", "TotalStorage BTU", derived_format_bold)
    w.write("E10", round(cf.BTU_PER_KWH * flo.max_energy_kwh_th), derived_format)

    w.write("D12", "EmitterPumpFeedbackModel", bold)
    w.write("E12", flo.params.EmitterPumpFeedbackModel.value)

    w.write("D13", "MixingValveFeedbackModel", bold)
    w.write("E13", flo.params.MixingValveFeedbackModel.value)

    w.write("D14", "IsRegulating", bold)
    w.write("E14", flo.params.IsRegulating)

    w.write("D19", "RoomTempF", bold)
    w.write("E19", flo.params.RoomTempF)

    w.write("D20", "AmbientPowerInKw", bold)
    w.write("E20", flo.params.AmbientPowerInKw)

    ###############

    w.write("G3", "HeatpumpTariff", bold)
    w.write("H3", flo.params.HeatpumpTariff.value)

    w.write("G4", "HeatpumpEnergySupplyType", bold)
    w.write("H4", flo.params.HeatpumpEnergySupplyType.value)

    w.write("G5", "BoostTariff", bold)
    w.write("H5", flo.params.BoostTariff.value)

    w.write("G6", "BoostEnergySupplyType", bold)
    w.write("H6", flo.params.BoostEnergySupplyType.value)

    w.write("G7", "StandardOfferPriceDollarsPerMwh", bold)
    w.write("H7", flo.params.StandardOfferPriceDollarsPerMwh)

    w.write("G8", "DistributionTariffDollarsPerMwh", bold)
    w.write("H8", flo.params.DistributionTariffDollarsPerMwh)

    w.write("A29", "WeatherUid", bold)
    w.write("B29", flo.params.WeatherUid)
    w.write("A31", "RtElecPriceUid", bold)
    w.write("B31", flo.params.RtElecPriceUid)

    w.write("A33", "DistPriceUid", bold)
    w.write("B33", flo.params.DistPriceUid)

    if flo.params.IsRegulating:
        w.write("A34", "LocalRegulationFile", bold)
        w.write("B34", regp_sync_100_handler.csv_file_by_uid(flo.params.RegPriceUid))
        w.write("A35", "RegPriceUid", bold)
        w.write("B35", flo.params.RegPriceUid)



""" SCADA Actor """
import functools
import logging
import time
from typing import Optional
from typing import cast

import dotenv
import pendulum
from gridworks.enums import GNodeRole
from pydantic import BaseModel

import gwatn.api_types as api_types
import gwatn.config as config
from gwatn.simple_scada_sim_actor_base import SimpleScadaSimActorBase
from gwatn.types import AtnParamsBrickstorageheater as AtnParams
from gwatn.types import AtnParamsBrickstorageheater_Maker
from gwatn.types import GtDispatchBoolean
from gwatn.types import SimplesimDriverDataBsh
from gwatn.types import SimplesimDriverDataBsh_Maker
from gwatn.types import SimplesimDriverReport
from gwatn.types import SimplesimSnapshotBrickstorageheater as Snapshot
from gwatn.types import SimTimestep


DISPATCH_CONTRACT_REPORTING_ALGOS = 5

LOG_FORMAT = (
    "%(levelname) -10s %(sasctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)

LOGGER.setLevel(logging.INFO)


def get_max_store_kwh_th(params: AtnParams) -> float:
    room_temp_c = (params.RoomTempF - 32) * 5 / 9
    return params.C * (params.MaxBrickTempC - room_temp_c)


class AtnHbStatus(BaseModel):
    LastHeartbeatReceivedMs: int
    AtnLastHex: Optional[str] = None
    ScadaLastHex: str = "0"


class SimpleScadaSim__BrickStorageHeater(SimpleScadaSimActorBase):
    def __init__(
        self,
        settings: config.ScadaSettings = config.ScadaSettings(
            _env_file=dotenv.find_dotenv()
        ),
    ):
        super().__init__(settings=settings)
        self.api_type_maker_by_name = api_types.TypeMakerByName
        self.atn_params_type_name = AtnParamsBrickstorageheater_Maker.type_name
        self.power_watts: float = 0
        self.store_kwh: int = 0
        self.max_store_kwh: int = 0
        self.atn_params: Optional[AtnParams] = None

    ########################
    ## Receives
    ########################

    #########################################################
    # Make the below into abstractmethods if pulling out base class
    #########################################################

    def simplesim_driver_report_received(self, payload: SimplesimDriverReport) -> None:
        """This gets received right before the top of the hour, from our
        best simulation of the TerminalAsset (which is happening in the
        AtomicTNode)."""
        if payload.FromGNodeInstanecId != self.atn_gni_id:
            LOGGER.info(f"Igoring {payload} - incorrect GNodeInstanceId")

        if payload.DriverDataTypeName != SimplesimDriverDataBsh_Maker.type_name:
            LOGGER.info(
                f"Igoring {payload} - incorrect DriverDataTypeName. "
                f"Expected {SimplesimDriverDataBsh_Maker.type_name} but got {payload.DriverDataTypeName}"
            )
        data = cast(SimplesimDriverDataBsh, payload.DriverData)
        self.power_watts = data.PowerWatts
        self.store_kwh = data.StoreKwh
        self.max_store_kwh = data.MaxStoreKwh
        self.send_snapshot()

    def send_snapshot(self):
        """Send a snapshot of current core sensed values to AtomicTNode.
        This is done every hour, and also on sensed power change."""
        if self.atn_params is None:
            return
        report_payload = Snapshot(
            FromGNodeAlias=self.alias,
            FromGNodeInstanceId=self.g_node_instance_id,
            PowerWatts=self.power_watts,
            StoreKwh=self.store_kwh,
            MaxStoreKwh=get_max_store_kwh_th(self.atn_params),
            AboutTerminalAssetAlias=self.ta_alias,
        )
        self.send_message(
            payload=report_payload,
            to_role=GNodeRole.AtomicTNode,
            to_g_node_alias=self.atn_alias,
        )

    def dispatch_received(self, payload: GtDispatchBoolean) -> None:
        """
        Dispatch received from AtomicTNode

          - Checks that the GNodeAlias and GNodeInstanceId belong to its
        AtomicTNode and that we have a dispatch contract
          - Sets talking_with to true
          - Follows instructions (turns on or turns off). Will turn on or
          off boost unless AboutNodeName is "a.element"

        For hourly sim:
          - Updatespower
          - Send status to AtomicTNode
        """
        if self.atn_params is None or self.in_dispatch_contract() is False:
            LOGGER.info("Igoring dispatch command, DispatchContract is not started")
        if payload.FromGNodeInstanceId != self.atn_gni_id:
            LOGGER.info(f"Igoring {payload}, not my Atn's GNodeInstanceId")
        self.talking_with = True
        if payload.AboutNodeName == "a.elements":
            # Making the grossly simplifying assumption that the heat pump turns on immediately
            if payload.RelayState == 1:
                new_power_watts = self.atn_params.RatedMaxPowerKw * 1000
            else:
                new_power_watts = 0
            if self.power_watts != new_power_watts:
                self.power_watts = new_power_watts
                self.send_snapshot()

    def new_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("New timestep")
        if not self.in_dispatch_contract():
            self.initialize_dispatch_contract()

    def repeat_timestep(self, payload: SimTimestep) -> None:
        LOGGER.info("Timestep received again")

    def time_str(self) -> str:
        return pendulum.from_timestamp(self.time()).strftime("%m/%d/%Y, %H:%M")

    def dispatch_contract_created(self) -> bool:
        """The SCADA only creates one kind of smart contract, which is its dispatch contract.
        So this just checks for a created app
        """
        created_apps = self.client.account_info(self.acct.addr)["created-apps"]
        if len(created_apps) > 1:
            raise NotImplementedError(
                f"SCADA not designed yet to create multiple dispatch contracts"
            )
        if len(created_apps) > 0:
            return True
        return False

    def in_dispatch_contract(self) -> bool:
        """Checks that bootstrap 2 was completed w addition of ta_trading_rights_idx,
        and also that Scada is opted in"""
        app_state = self.dc_client.get_application_state()
        if "ta_trading_rights_idx" not in app_state.keys():
            LOGGER.info(
                "Not in Dispatch Contract because ta_trading_rights_idx isn't in app_state"
            )
            LOGGER.info(
                "AtomicTNode has not finished its part in bootstrapping the contract"
            )
            return False
        # Todo: also check that atn is opted in
        apps = self.client.account_info(self.acct.addr)["apps-local-state"]
        app_ids = list(map(lambda x: x["id"], apps))
        if self.dc_client.app_id in app_ids:
            return True
        LOGGER.info(
            "Not in Dispatch Contract because scada has not yet opted into contract"
        )
        return False



# type: ignore
import math
import random
import time
from typing import Dict
from typing import List
from typing import Optional
from typing import no_type_check

import numpy as np
import pendulum

import gwatn.brick_storage_heater.strategy_utils as strategy_utils
from gwatn.brick_storage_heater.edge import Edge__BrickStorageHeater as Edge
from gwatn.brick_storage_heater.node import Node_BrickStorageHeater as Node
from gwatn.data_classes.d_graph import DGraph
from gwatn.types import FloParamsBrickstorageheater as FloParams


class Flo__BrickStorageHeater(DGraph):
    MAGIC_HEAT_PUMP_DELTA_F = 20
    FAILED_HEATING_PENALTY_DOLLARS = 10**6

    def __init__(
        self,
        params: FloParams,
        d_graph_id: str,
    ):
        self.params = params

        self.RealtimeElectricityPrice = np.array(self.params.RealtimeElectricityPrice)
        self.DistributionPrice = np.array(self.params.DistributionPrice)
        if self.params.IsRegulating:
            self.reg_price_per_mwh = np.array(self.params.RegulationPrice)
        else:
            self.reg_price_per_mwh = np.array(
                [0] * len(self.params.RealtimeElectricityPrice)
            )

        self.max_energy_kwh_th = strategy_utils.get_max_store_kwh_th(self.params)

        self.currency_unit = self.params.CurrencyUnit
        self.temp_unit = self.params.TempUnit
        DGraph.__init__(
            self,
            d_graph_id=d_graph_id,
            graph_strategy_alias="SpaceHeat__HeatPumpWithBoostStore__Flo",
            flo_start_unix_time_s=pendulum.datetime(
                year=self.params.StartYearUtc,
                month=self.params.StartMonthUtc,
                day=self.params.StartDayUtc,
                hour=self.params.StartHourUtc,
                minute=self.params.StartMinuteUtc,
            ).timestamp(),
            slice_duration_hrs=list(np.array(self.params.SliceDurationMinutes) / 60),
            default_storage_steps=self.params.StorageSteps,
            starting_store_idx=self.params.StartingStoreIdx,
            timezone_string=self.params.TimezoneString,
            home_city=self.params.HomeCity,
            currency_unit=self.params.CurrencyUnit,
            max_storage=self.max_energy_kwh_th,
            max_power_in=self.params.RatedMaxPowerKw,
            wh_exponent=3,
        )
        self.e_step: float = self.max_energy_kwh_th / self.params.StorageSteps
        room_temp_c = (self.params.RoomTempF - 32) * 5 / 9
        temp_range_c = self.params.MaxBrickTempC - room_temp_c
        self.temp_step_c: float = temp_range_c / self.params.StorageSteps
        self.store_kwh_per_deg_c = self.max_energy_kwh_th / temp_range_c
        self.energy_cost_per_kwh: Dict[int, float] = {}
        self.set_energy_cost_per_kwh()
        self.uncosted_edges: Dict[Edge, int] = {}
        self.failed_in_cost_boost_preferred: int = 0
        self.failed_in_cost_hp_preferred: int = 0

        self.create_graph()
        self.solve_dijkstra()

    def create_slice_nodes(self, ts_idx: int) -> None:
        """Creates nodes for time slice ts_idx, equally spaced by self.e_step,
        running from params.ZeroPotentialEnergyWaterTempF up to params.MaxStoreTempF.

        Sets the energy store in kwh of enthalpy (latent heat in this case) as well
        as the average boost water temp.

        """
        ...

    def create_nodes(self) -> None:
        for ts_idx in range(self.time_slices + 1):
            self.create_slice_nodes(ts_idx)

    def set_energy_cost_per_kwh(self):
        for ts_idx in range(self.time_slices):
            self.energy_cost_per_kwh[ts_idx] = self.get_energy_cost_per_kwh(ts_idx)

    ######################################################
    # Functions of edges
    ######################################################

    # Cost related

    def get_energy_cost_per_kwh(self, ts_idx: int) -> float:
        energy_cost_per_mwh = self.params.RealtimeElectricityPrice[ts_idx]
        dist_cost_per_mwh = self.params.DistributionPrice[ts_idx]

        cost_per_mwh = energy_cost_per_mwh + dist_cost_per_mwh
        return cost_per_mwh / 1000

    ######################################################
    # Functions of edges - Water temp related
    #####################################################

    def set_failed_edge_properties(self, edge: Edge):
        """Set costs"""
        edge.cost = self.FAILED_HEATING_PENALTY_DOLLARS
        edge.hp_electricity_avg_kw = 10000
        edge.avg_kw = 0

    def get_failed_edge(self, node: Node) -> Edge:
        edge = Edge(
            start_ts_idx=node.ts_idx, start_idx=node.store_idx, end_idx=node.store_idx
        )
        self.set_failed_edge_properties(edge)
        return edge

    def get_uncosted_edge(
        self, node: Node, delta_energy_kwh: float, existing_edges: List[Edge] = []
    ) -> Optional[Edge]:
        """Finds a randomized passing edge consistent with delta_energy_kwh.
        For example, if delta_energy_kwh falls exactly half way between
        two energy steps it will return each step with probability 50%
        """
        if node.store_enthalpy_kwh + delta_energy_kwh < 0:
            return None
        delta_idx_as_float = delta_energy_kwh / self.e_step
        lower_delta_idx = math.floor(delta_idx_as_float)
        frac = delta_idx_as_float - lower_delta_idx
        random_adder = random.choices(population=[0, 1], weights=[1 - frac, frac], k=1)[
            0
        ]
        if node.store_idx + lower_delta_idx == self.params.StorageSteps:
            delta_idx = lower_delta_idx
        else:
            delta_idx = lower_delta_idx + random_adder
        edge = Edge(
            start_ts_idx=node.ts_idx,
            start_idx=node.store_idx,
            end_idx=node.store_idx + delta_idx,
        )
        if edge not in existing_edges:
            self.uncosted_edges[edge] = 1
            return edge
        else:
            # in this case the edge is already in the existing_edges, so return None
            return None

    def get_edges_from_node(self, node: Node) -> List[Edge]:
        """Creates edges starting at a node."""
        ts_idx = node.ts_idx
        return []

    def set_edge_cost(self, edge: Edge) -> None:
        """Given an edge:
        - set cost to penalty if there is no way to meet the energy requirements
        - otherwise set the cost for the optimal choice of using the boost and
        the heat pump
        """
        ...

    def create_graph(self) -> None:
        # print(f"Creating graph nodes, edges and edge weights")
        st = time.time()
        self.create_nodes()
        nt = time.time()
        # print(f"{time.time() - st:1.0f} seconds to make nodes")
        for jj in range(self.time_slices):
            # if jj % 500 == 0:
            #     print(f"{time.time() - st:1.0f} seconds for {jj} slices")
            for node in self.node[jj].values():
                edges = self.get_edges_from_node(node=node)
                self.edges[node] = edges
        et = time.time()
        # print(f"{et - nt:1.0f} seconds for building edges")
        for edge in self.uncosted_edges.keys():
            self.set_edge_cost(edge)
        self.uncosted_edges = {}
        ct = time.time()
        # print(f"{ct - et:1.0f} seconds for costing edges")
        tt = time.time() - st
        # print(f"time per 100 slices:{tt*100000/self.time_slices:2.0f} ms ")
        # print(f"total time to build graph: {tt:2.0f} s")
        # print(f"Total number of slices: {self.time_slices}")



import csv
import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple

import gridworks.property_format as property_format
import numpy as np
from gridworks.errors import SchemaError


class CsvEprtSync(NamedTuple):
    TypeName: str
    PNodeAlias: str
    MethodAlias: str
    Comment: str
    StartYearUtc: int
    StartMonthUtc: int
    StartDayUtc: int
    StartHourUtc: int
    StartMinuteUtc: int
    UniformSliceDurationHrs: float
    TimezoneString: str
    CurrencyUnit: str
    PriceUid: str
    Header: str
    Prices: list

    def asdict(self):
        d = self._asdict()
        return d

    def is_valid(self) -> Tuple[bool, Optional[List[str]]]:
        is_valid, errors = self.passes_derived_validations()
        return is_valid, errors

    def passes_derived_validations(self) -> Tuple[bool, Optional[List[str]]]:
        is_valid = True
        errors = []
        if self.TypeName != "csv.eprt.sync":
            is_valid = False
            errors.append(
                f"Payload requires TypeName of 'csv.eprt.sync', not {self.TypeName}."
            )
        if not isinstance(self.PNodeAlias, str):
            is_valid = False
            errors.append(f"PNodeAlias must have type str.")
        if not isinstance(self.StartYearUtc, int):
            is_valid = False
            errors.append(f"StartYearUtc must have type int.")
        if not isinstance(self.StartMonthUtc, int):
            is_valid = False
            errors.append(f"StartMonthUtc must have type int.")
        if not isinstance(self.StartDayUtc, int):
            is_valid = False
            errors.append(f"StartDayUtc must have type int.")
        if not isinstance(self.StartHourUtc, int):
            is_valid = False
            errors.append(f"StartHourUtc must have type int.")
        if not isinstance(self.StartMinuteUtc, int):
            is_valid = False
            errors.append(f"StartMinuteUtc must have type int.")
        try:
            datetime.datetime(
                year=self.StartYearUtc,
                month=self.StartMonthUtc,
                day=self.StartDayUtc,
                hour=self.StartHourUtc,
                minute=self.StartMinuteUtc,
            )
        except ValueError as e:
            is_valid = False
            errors.append(e)
        if not isinstance(self.UniformSliceDurationHrs, float):
            is_valid = False
            errors.append(f"UniformSliceDurationHrs must have type float.")
        if not isinstance(self.TimezoneString, str):
            is_valid = False
            errors.append(f"TimezoneString {self.TimezoneString} must have type str.")
        if not isinstance(self.CurrencyUnit, str):
            is_valid = False
            errors.append(f"CurrencyUnit {self.CurrencyUnit} must have type str.")
        if not isinstance(self.PriceUid, str):
            is_valid = False
            errors.append(f"PriceUid {self.PriceUid} must have type str.")
        if not property_format.is_uuid_canonical_textual(self.PriceUid):
            is_valid = False
            errors.append(
                f"PriceUid {self.PriceUid} must have format UuidCanonicalTextual."
            )

        return is_valid, errors


class CsvEprtSync_Maker:
    type_name = "csv.eprt.sync"

    @classmethod
    def payload_is_valid(
        cls, payload_as_dict: Dict[str, Any]
    ) -> Tuple[bool, Optional[List[str]]]:
        try:
            payload_as_dict["UniformSliceDurationHrs"] = float(
                payload_as_dict["UniformSliceDurationHrs"]
            )
        except ValueError:
            pass  # This will get caught in is_valid() check below
        try:
            p = CsvEprtSync(
                TypeName=payload_as_dict["TypeName"],
                PNodeAlias=payload_as_dict["PNodeAlias"],
                MethodAlias=payload_as_dict["MethodAlias"],
                Comment=payload_as_dict["Comment"],
                StartYearUtc=payload_as_dict["StartYearUtc"],
                StartMonthUtc=payload_as_dict["StartMonthUtc"],
                StartDayUtc=payload_as_dict["StartDayUtc"],
                StartHourUtc=payload_as_dict["StartHourUtc"],
                StartMinuteUtc=payload_as_dict["StartMinuteUtc"],
                UniformSliceDurationHrs=payload_as_dict["UniformSliceDurationHrs"],
                TimezoneString=payload_as_dict["TimezoneString"],
                CurrencyUnit=payload_as_dict["CurrencyUnit"],
                PriceUid=payload_as_dict["PriceUid"],
                Header=payload_as_dict["Header"],
                Prices=payload_as_dict["Prices"],
            )
        except TypeError:
            errors = [TypeError]
            return False, errors
        return p.is_valid()

    def __init__(self, elec_price_file):
        file = elec_price_file
        self.errors = []
        self.tuple: Optional[CsvEprtSync] = None
        first_data_row = 14
        prices = np.array([])
        uniform_slice_duration_hrs = np.array([])
        with open(elec_price_file, newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=",")
            line_idx = 0
            for row in reader:
                if line_idx == 0:
                    property = row[0].replace("\ufeff", "")
                    mp_alias = row[1]
                    if property != "TypeName":
                        raise Exception(
                            f"Csv 0,0 must be TypeName, instead it is {property}"
                        )
                if line_idx == 1:
                    property = row[0]
                    p_node_alias = row[1].strip()
                    if property != "PNodeAlias":
                        raise Exception(
                            f"Csv 1,0 must be PNodeAlias, instead it is {property}"
                        )
                if line_idx == 2:
                    property = row[0]
                    method_alias = row[1].strip()
                    if property != "MethodAlias":
                        raise Exception(
                            f"Csv 2,0 must be MethodAlias, instead it is {property}. Check file {file}."
                        )
                if line_idx == 3:
                    property = row[0]
                    comment = row[1].strip()
                    if property != "Comment":
                        raise Exception(
                            f"Csv 3,0 must be Comment, instead it is {property}"
                        )
                if line_idx == 4:
                    property = row[0]
                    start_year_utc = int(row[1])
                    if property != "StartYearUtc":
                        raise Exception(
                            f"Csv 4,0 must be StartYearUtc, instead it is {property}"
                        )
                if line_idx == 5:
                    property = row[0]
                    start_month_utc = int(row[1])
                    if property != "StartMonthUtc":
                        raise Exception(
                            f"Csv 5,0 must be StartMonthUtc, instead it is {property}"
                        )
                if line_idx == 6:
                    property = row[0]
                    start_day_utc = int(row[1])
                    if property != "StartDayUtc":
                        raise Exception(
                            f"Csv 6,0 must be StartDayUtc, instead it is {property}"
                        )
                if line_idx == 7:
                    property = row[0]
                    start_hour_utc = int(row[1])
                    if property != "StartHourUtc":
                        raise Exception(
                            f"Csv 7,0 must be StartHourUtc, instead it is {property}"
                        )
                if line_idx == 8:
                    property = row[0]
                    start_minute_utc = int(row[1])
                    if property != "StartMinuteUtc":
                        raise Exception(
                            f"Csv 8,0 must be StartMinuteUtc, instead it is {property}"
                        )
                if line_idx == 9:
                    property = row[0]
                    uniform_slice_duration_hrs = float(row[1])
                    if property != "UniformSliceDurationHrs":
                        raise Exception(
                            f"Csv 9,0 must be UniformSliceDurationHrs, instead it is {property}"
                        )
                if line_idx == 10:
                    property = row[0]
                    timezone_string = row[1].strip()
                    if property != "TimezoneString":
                        raise Exception(
                            f"Csv 10,0 must be TimezoneString, instead it is {property}"
                        )
                if line_idx == 11:
                    property = row[0]
                    currency_unit = row[1].strip()
                    if property != "CurrencyUnit":
                        raise Exception(
                            f"Csv 11,0 must be CurrencyUnit, instead it is {property}"
                        )
                if line_idx == 12:
                    property = row[0]
                    price_uid = row[1].strip()
                    if property != "PriceUid":
                        raise Exception(
                            f"Csv 12,0 must be PriceUid, instead it is {property}"
                        )
                if line_idx == 13:
                    property = row[0]
                    header = row[1].strip()
                    if property != "Header":
                        raise Exception(
                            f"Csv 13,0 must be Header, instead it is {property}"
                        )
                    if header != "Real Time LMP Electricity Price (Currency Unit/MWh)":
                        raise Exception(
                            f"Csv 13,1 must be 'Real Time LMP Electricity Price (Currency Unit/MWh)', instead it is {header}"
                        )
                elif line_idx >= first_data_row:
                    try:
                        prices = np.append(prices, float(row[0]))
                    except ValueError:
                        raise Exception(f"Missing a price in row {line_idx+1}")
                line_idx += 1

        p = CsvEprtSync(
            TypeName=mp_alias,
            PNodeAlias=p_node_alias,
            MethodAlias=method_alias,
            Comment=comment,
            StartYearUtc=start_year_utc,
            StartMonthUtc=start_month_utc,
            StartDayUtc=start_day_utc,
            StartHourUtc=start_hour_utc,
            StartMinuteUtc=start_minute_utc,
            UniformSliceDurationHrs=uniform_slice_duration_hrs,
            TimezoneString=timezone_string,
            CurrencyUnit=currency_unit,
            PriceUid=price_uid,
            Header=header,
            Prices=list(prices),
        )

        is_valid, errors = p.is_valid()
        if is_valid is False:
            self.errors = errors
            print(errors)
            raise SchemaError(
                f"Failed to create payload due to these errors:{errors}. Input file is {elec_price_file}"
            )
        else:
            self.tuple = p



"""Type accepted.bid, version 000"""
import json
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import validator

from gwatn.types.price_quantity_unitless import PriceQuantityUnitless
from gwatn.types.price_quantity_unitless import PriceQuantityUnitless_Maker


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_market_type_name_lrd_format(v: str) -> None:
    from gwatn.enums import MarketTypeName

    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    if not x[0] in MarketTypeName.values():
        raise ValueError(f"{v} not recognized MarketType")
    g_node_alias = ".".join(x[1:])
    check_is_left_right_dot(g_node_alias)


def check_is_market_slot_name_lrd_format(v: str) -> None:
    """
    MaketSlotNameLrdFormat: the format of a MarketSlotName.
      - The first word must be a MarketTypeName
      - The last word (unix time of market slot start) must
      be a 10-digit integer divisible by 300 (i.e. all MarketSlots
      start at the top of 5 minutes)
      - More strictly, the last word must be the start of a
      MarketSlot for that MarketType (i.e. divisible by 3600
      for hourly markets)
      - The middle words have LeftRightDot format (GNodeAlias
      of the MarketMaker)
    Example: rt60gate5.d1.isone.ver.keene.1673539200

    """
    from gwatn.data_classes import MarketType

    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    slot_start = x[-1]
    if len(slot_start) != 10:
        raise ValueError(f"slot start {slot_start} not of length 10")
    try:
        slot_start = int(slot_start)
    except ValueError:
        raise ValueError(f"slot start {slot_start} not an int")
    if slot_start % 300 != 0:
        raise ValueError(f"slot start {slot_start} not a multiple of 300")

    market_type_name_lrd = ".".join(x[:-1])
    try:
        check_is_market_type_name_lrd_format(market_type_name_lrd)
    except ValueError as e:
        raise ValueError(f"e")

    market_type = MarketType.by_id[market_type_name_lrd.split(".")[0]]
    if not slot_start % (market_type.duration_minutes * 60) == 0:
        raise ValueError(
            f"market_slot_start_s mod {(market_type.duration_minutes * 60)} must be 0"
        )


class AcceptedBid(BaseModel):
    """Bid acceptance sent from MarketMaker to a market partipant.

    This is a legally binding contract for the bidder to consume or produce the quantity
    in its Bid consistent with the actual price.
    [More info](https://gridworks.readthedocs.io/en/latest/market-bid.html).
    """

    MarketSlotName: str = Field(
        title="MarketSlotName",
    )
    BidderAlias: str = Field(
        title="BidderAlias",
    )
    PqPairs: List[PriceQuantityUnitless] = Field(
        title="PqPairs",
    )
    ReceivedTimeUnixNs: int = Field(
        title="ReceivedTimeUnixNs",
    )
    TypeName: Literal["accepted.bid"] = "accepted.bid"
    Version: str = "000"

    @validator("MarketSlotName")
    def _check_market_slot_name(cls, v: str) -> str:
        try:
            check_is_market_slot_name_lrd_format(v)
        except ValueError as e:
            raise ValueError(
                f"MarketSlotName failed MarketSlotNameLrdFormat format validation: {e}"
            )
        return v

    @validator("BidderAlias")
    def _check_bidder_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"BidderAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("PqPairs")
    def _check_pq_pairs(cls, v: List) -> List:
        for elt in v:
            if not isinstance(elt, PriceQuantityUnitless):
                raise ValueError(
                    f"elt {elt} of PqPairs must have type PriceQuantityUnitless."
                )
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()

        # Recursively call as_dict() for the SubTypes
        pq_pairs = []
        for elt in self.PqPairs:
            pq_pairs.append(elt.as_dict())
        d["PqPairs"] = pq_pairs
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class AcceptedBid_Maker:
    type_name = "accepted.bid"
    version = "000"

    def __init__(
        self,
        market_slot_name: str,
        bidder_alias: str,
        pq_pairs: List[PriceQuantityUnitless],
        received_time_unix_ns: int,
    ):
        self.tuple = AcceptedBid(
            MarketSlotName=market_slot_name,
            BidderAlias=bidder_alias,
            PqPairs=pq_pairs,
            ReceivedTimeUnixNs=received_time_unix_ns,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: AcceptedBid) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> AcceptedBid:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> AcceptedBid:
        d2 = dict(d)
        if "MarketSlotName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MarketSlotName")
        if "BidderAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing BidderAlias")
        if "PqPairs" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PqPairs")
        pq_pairs = []
        if not isinstance(d2["PqPairs"], List):
            raise SchemaError("PqPairs must be a List!")
        for elt in d2["PqPairs"]:
            if not isinstance(elt, dict):
                raise SchemaError(
                    f"elt {elt} of PqPairs must be "
                    "PriceQuantityUnitless but not even a dict!"
                )
            pq_pairs.append(PriceQuantityUnitless_Maker.dict_to_tuple(elt))
        d2["PqPairs"] = pq_pairs
        if "ReceivedTimeUnixNs" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ReceivedTimeUnixNs")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return AcceptedBid(
            MarketSlotName=d2["MarketSlotName"],
            BidderAlias=d2["BidderAlias"],
            PqPairs=d2["PqPairs"],
            ReceivedTimeUnixNs=d2["ReceivedTimeUnixNs"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type tavalidatorcert.algo.transfer, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class TavalidatorcertAlgoTransfer(BaseModel):
    """Used for Step 2 of TaValidator certification.

    Meant to be sent from a pending TaValidator  to the GNodeFactory (Gnf), so the
    Gnf will transfer its ValidatorCert to the pending TaValidator's Algorand address.
    [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html).
    """

    ValidatorAddr: str = Field(
        title="The address of the pending TaValidator",
    )
    HalfSignedCertTransferMtx: str = Field(
        title="Algo multi-transaction for certificate transfer, with 1 of 2 signatures",
    )
    TypeName: Literal["tavalidatorcert.algo.transfer"] = "tavalidatorcert.algo.transfer"
    Version: str = "000"

    @validator("ValidatorAddr")
    def check_validator_addr(cls, v: str) -> str:
        """
        Axiom 4: TaValidator has sufficient Algos.
        MultiAccount [GnfAdminAddr, ValidatorAddr] must have enough Algos to meet
        the GNodeFactory criterion.
        """
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        # TODO: Axiom
        return v

    @validator("HalfSignedCertTransferMtx")
    def _check_half_signed_cert_transfer_mtx(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"HalfSignedCertTransferMtx failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator(pre=True)
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Is correct Multisig.
        Decoded HalfSignedCertTransferMtx must have type MultisigTransaction from the
        2-sig MultiAccount  [GnfAdminAddr, ValidatorAddr], signed by the ValidatorAddr
        [More info](https://gridworks.readthedocs.io/en/latest/g-node-factory.html#gnfadminaddr)
        """
        # TODO: Implement check for axiom 1"
        return v

    @root_validator
    def check_axiom_2(cls, v: dict) -> dict:
        """
        Axiom 2: Transfers correct certificate.
         - The transaction must be the transfer of an Algorand Standard Asset
         -The sender must be the 2-sig Multi [GnfAdminAddr, TaValidatorAddr], which also created and owns the ASA
         - It must be getting sent to the ValidatorAddr
         -The ASA must have:
           - Total = 1
           - UnitName=VLDITR
           - GnfAdminAddr as manage
           - AssetName not blank
        - The transfer amount must be 1
        [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate)
        """
        # TODO: Implement check for axiom 2"
        return v

    @root_validator
    def check_axiom_3(cls, v: dict) -> dict:
        """
        Axiom 3: TaValidator has opted in.
        ValidatorAddr must be opted into the transferring ASA.
        """
        # TODO: Implement check for axiom 3"
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class TavalidatorcertAlgoTransfer_Maker:
    type_name = "tavalidatorcert.algo.transfer"
    version = "000"

    def __init__(self, validator_addr: str, half_signed_cert_transfer_mtx: str):
        self.tuple = TavalidatorcertAlgoTransfer(
            ValidatorAddr=validator_addr,
            HalfSignedCertTransferMtx=half_signed_cert_transfer_mtx,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: TavalidatorcertAlgoTransfer) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> TavalidatorcertAlgoTransfer:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> TavalidatorcertAlgoTransfer:
        d2 = dict(d)
        if "ValidatorAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValidatorAddr")
        if "HalfSignedCertTransferMtx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HalfSignedCertTransferMtx")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return TavalidatorcertAlgoTransfer(
            ValidatorAddr=d2["ValidatorAddr"],
            HalfSignedCertTransferMtx=d2["HalfSignedCertTransferMtx"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type tavalidatorcert.algo.create, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class TavalidatorcertAlgoCreate(BaseModel):
    """Used for Step 1 of TaValidator certification.

    Meant to be sent from a pending TaValidator  to the GNodeFactory (Gnf), to
    initiate the process of certifying the pending TaValidator.
    [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html).
    """

    ValidatorAddr: str = Field(
        title="The address of the pending TaValidator",
    )
    HalfSignedCertCreationMtx: str = Field(
        title="Algo multi-transaction for certificate creation, with 1 of 2 signatures",
    )
    TypeName: Literal["tavalidatorcert.algo.create"] = "tavalidatorcert.algo.create"
    Version: str = "000"

    @validator("ValidatorAddr")
    def check_validator_addr(cls, v: str) -> str:
        """
        Axiom 5: Uniqueness.
        There must not already be a TaValidatorCert belonging to the  2-sig [GnfAdminAddr, ValidatorAddr] address.
        """
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ValidatorAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        # TODO: Axiom
        return v

    @validator("HalfSignedCertCreationMtx")
    def check_half_signed_cert_creation_mtx(cls, v: str) -> str:
        """
        Axioms 2, 3:

        Axiom 2: Is AssetConfigTxn.
        The transaction must have type AssetConfigTxn.

        Axiom 3: Is ValidatorCert.
        For the asset getting created: Total is 1, Decimals is 0, UnitName is VLDTR, Manager is GnfAdminAddr,
        AssetName is not blank.
        [More info](https://gridworks.readthedocs.io/en/latest/ta-validator.html#tavalidator-certificate)
        """
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"HalfSignedCertCreationMtx failed AlgoMsgPackEncoded format validation: {e}"
            )
        # TODO: Axiom
        return v

    @root_validator(pre=True)
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: Is correct Multisig.
        Decoded HalfSignedCertCreationMtx must have type MultisigTransaction from the
        2-sig MultiAccount  [GnfAdminAddr, ValidatorAddr], signed by ValidatorAddr.
        [More info](https://gridworks.readthedocs.io/en/latest/g-node-factory.html#gnfadminaddr)
        """
        # TODO: Implement check for axiom 1"
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class TavalidatorcertAlgoCreate_Maker:
    type_name = "tavalidatorcert.algo.create"
    version = "000"

    def __init__(self, validator_addr: str, half_signed_cert_creation_mtx: str):
        self.tuple = TavalidatorcertAlgoCreate(
            ValidatorAddr=validator_addr,
            HalfSignedCertCreationMtx=half_signed_cert_creation_mtx,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: TavalidatorcertAlgoCreate) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> TavalidatorcertAlgoCreate:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> TavalidatorcertAlgoCreate:
        d2 = dict(d)
        if "ValidatorAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ValidatorAddr")
        if "HalfSignedCertCreationMtx" not in d2.keys():
            raise SchemaError(f"dict {d2} missing HalfSignedCertCreationMtx")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return TavalidatorcertAlgoCreate(
            ValidatorAddr=d2["ValidatorAddr"],
            HalfSignedCertCreationMtx=d2["HalfSignedCertCreationMtx"],
            TypeName=d2["TypeName"],
            Version="000",
        )



""" AtnActorBase """
import functools
import logging
import random
import time
import uuid
from abc import abstractmethod
from typing import Optional
from typing import cast
from typing import no_type_check

import gridworks.algo_utils as algo_utils
import pendulum
from algosdk.atomic_transaction_composer import TransactionWithSigner
from algosdk.future.transaction import *
from algosdk.v2client.algod import AlgodClient
from beaker.client import ApplicationClient
from gridworks.algo_utils import BasicAccount
from gridworks.enums import GNodeRole
from gridworks.message import as_enum
from gwproto.message import Message as ScadaMessage
from gwproto.messages import Ack
from gwproto.messages import GtShStatusEvent
from gwproto.messages import PeerActiveEvent
from gwproto.messages import Ping as GridworksPing
from gwproto.messages import SnapshotSpaceheatEvent
from pydantic import BaseModel

from gwatn.config import AtnSettings
from gwatn.dispatch_contract import DispatchContract
from gwatn.enums import AlgoCertType
from gwatn.enums import MessageCategorySymbol
from gwatn.enums import UniverseType
from gwatn.two_channel_actor_base import TwoChannelActorBase
from gwatn.types import AtnParams
from gwatn.types import DispatchContractConfirmed_Maker
from gwatn.types import GtShStatus
from gwatn.types import GwCertId
from gwatn.types import GwCertId_Maker
from gwatn.types import HeartbeatA
from gwatn.types import HeartbeatA_Maker
from gwatn.types import HeartbeatB
from gwatn.types import HeartbeatB_Maker
from gwatn.types import JoinDispatchContract
from gwatn.types import JoinDispatchContract_Maker
from gwatn.types import LatestPrice
from gwatn.types import LatestPrice_Maker
from gwatn.types import PowerWatts
from gwatn.types import SimTimestep
from gwatn.types import SimTimestep_Maker
from gwatn.types import SnapshotSpaceheat


LOG_FORMAT = (
    "%(levelname) -10s %(sasctime)s %(name) -30s %(funcName) "
    "-35s %(lineno) -5d: %(message)s"
)
LOGGER = logging.getLogger(__name__)


class HbStatus(BaseModel):
    LastHeartbeatReceivedMs: int
    AtnLastHex: str = "0"
    ScadaLastHex: str = "0"
    # To see beaker issue an ABI complaint:
    # ScadaLastHex: Optional[str] = None
    #
    # This will create an HbStatus with ScadaLastHex of None
    # when the Atn is initialized.


def dummy_atn_params() -> AtnParams:
    return AtnParams(
        SliceDurationMinutes=60,
        FloSlices=48,
        GNodeAlias="d1.isone.dummy.ta",
        GNodeInstanceId="00000000-0000-0000-0000-000000000000",
        TypeName="atn.params.heatpumpwithbooststore",
        Version="000",
    )


class AtnActorBase(TwoChannelActorBase):
    def __init__(self, settings: AtnSettings, use_algo: bool = False):
        super().__init__(settings=settings)
        self.settings: AtnSettings = settings
        self.scada_gni_id = settings.scada_gni_id
        self._time: float = self.get_initial_time_s()

        if use_algo is True:
            self.acct: BasicAccount = BasicAccount(settings.sk.get_secret_value())
            self.client: AlgodClient = AlgodClient(
                settings.algo_api_secrets.algod_token.get_secret_value(),
                settings.public.algod_address,
            )
            if algo_utils.algos(self.acct.addr) < 5:
                raise Exception(
                    f"Insufficiently funded. Make sure atn has at least 5 algos"
                )
            # TODO: move this into spaceheat along with join_dispatch_contract_received
            self.atn_params: AtnParams = dummy_atn_params()
            self.sp = self.client.suggested_params()
            self.sp.flat_fee = True
            self.sp.fee = 2000
            # this is initialized with the AppId provided by the SCADA
            self.dc_app_id: Optional[int] = None
            self.dc_client: Optional[ApplicationClient] = None
            self.check_for_dispatch_contract()
            self.universe_type = as_enum(
                self.settings.universe_type_value, UniverseType, UniverseType.default()
            )
            self.trading_rights_id: Optional[GwCertId] = None
            self.update_trading_rights()
            self.hb_status = HbStatus(LastHeartbeatReceivedMs=int(time.time() * 1000))

    def local_rabbit_startup(self) -> None:
        rjb = MessageCategorySymbol.rjb.value
        tc_alias_lrh = self.settings.time_coordinator_alias.replace(".", "-")
        binding = f"{rjb}.{tc_alias_lrh}.timecoordinator.sim-timestep"

        cb = functools.partial(self.on_timecoordinator_bindok, binding=binding)
        self._consume_channel.queue_bind(
            self.queue_name, "timecoordinatormic_tx", routing_key=binding, callback=cb
        )
        LOGGER.info(
            f"Queue {self.queue_name} bound to timecoordinatormic_tx with {binding} "
        )
        scada_alias_lrh = self.scada_alias.replace(".", "-")
        binding = f"gw.{scada_alias_lrh}.#"
        cb = functools.partial(self.on_scada_bindok, binding=binding)
        self._consume_channel.queue_bind(
            self.queue_name, "amq.topic", routing_key=binding, callback=cb
        )
        LOGGER.info(f"Queue {self.queue_name} bound to amq.topic with {binding} ")
        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex="0"
        ).tuple
        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )
        d = pendulum.from_timestamp(time.time())
        LOGGER.warning(
            f"[{self.alias}] Sent first heartbeat to super: {d.minute}:{d.second}.{d.microsecond}"
        )
        self.strategy_rabbit_startup()

    def strategy_rabbit_startup(self) -> None:
        pass

    @no_type_check
    def on_timecoordinator_bindok(self, _unused_frame, binding) -> None:
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")

    @no_type_check
    def on_scada_bindok(self, _unused_frame, binding) -> None:
        LOGGER.info(f"Queue {self.queue_name} bound with {binding}")

    def time(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self._time
        else:
            return time.time()

    def get_initial_time_s(self) -> float:
        if self.universe_type == UniverseType.Dev:
            return self.settings.initial_time_unix_s
        else:
            return time.time()

    def prepare_for_death(self) -> None:
        self.actor_main_stopped = True

    ########################
    ## Receives
    ########################

    def route_scada_message(self, from_alias: str, message: ScadaMessage) -> None:
        """Routes messages from the SCADA to methods that should be overwritten
        in derived class"""

        # if message.Header.Src != self.scada_alias:
        #     LOGGER.info(
        #         f"Ignoring scada message from {message.Header.Src}. My scada is {self.scada_alias}"
        #     )
        #     LOGGER.info(f"{message}")
        #     return
        if message.Header.AckRequired == True:
            payload = Ack(AckMessageID=message.Header.MessageId)
            self.send_scada_message(payload)

        path_dbg = 0
        match message.Payload:
            case GridworksPing():
                self._process_gridworks_ping_from_scada(
                    cast(GridworksPing, message.Payload)
                )
            case GtShStatusEvent():
                self._process_gt_sh_status_event_from_scada(
                    cast(GtShStatusEvent, message.Payload)
                )
            case PeerActiveEvent():
                self._process_peer_active_event_from_scada(
                    cast(PeerActiveEvent, message.Payload)
                )
            case SnapshotSpaceheatEvent():
                self._process_snapshot_spaceheat_event_from_scada(
                    cast(SnapshotSpaceheatEvent, message.Payload)
                )
            case PowerWatts():
                self._process_power_watts_from_scada(cast(PowerWatts, message.Payload))

            case SnapshotSpaceheat():
                self._process_snapshot_spaceheat_from_scada(
                    cast(SnapshotSpaceheat, message.Payload)
                )

    @abstractmethod
    def _process_gridworks_ping_from_scada(self, payload: GridworksPing) -> None:
        """Atn has received gridworks.ping message from its SCADA"""
        raise NotImplementedError

    @abstractmethod
    def _process_peer_active_event_from_scada(self, payload: PeerActiveEvent) -> None:
        """Atn has received gridworks.event.comm.peer.active message from its SCADA"""
        raise NotImplementedError

    def _process_gt_sh_status_event_from_scada(self, payload: GtShStatusEvent) -> None:
        """Atn has received gridworks.event.gt.sh.status message from its SCADA"""
        gt_sh_status = payload.status
        self._process_gt_sh_status_from_scada(payload=gt_sh_status)

    def _process_snapshot_spaceheat_event_from_scada(
        self, payload: SnapshotSpaceheatEvent
    ) -> None:
        """Atn has received  gridworks.event.snapshot.spaceheat message from its SCADA"""
        snapshot_spaceheat = payload.snap
        self._process_snapshot_spaceheat_from_scada(payload=snapshot_spaceheat)

    @abstractmethod
    def _process_gt_sh_status_from_scada(self, payload: GtShStatus) -> None:
        """Atn has received gt.sh.status message from its SCADA"""
        raise NotImplementedError

    @abstractmethod
    def _process_power_watts_from_scada(self, payload: PowerWatts) -> None:
        """Atn has received power.watts message from its SCADA"""
        raise NotImplementedError

    @abstractmethod
    def _process_snapshot_spaceheat_from_scada(self, payload: SnapshotSpaceheat):
        """Atn has received a snapshot.sspaceheat message from the SCADA"""
        raise NotImplementedError

    def route_message(
        self, from_alias: str, from_role: GNodeRole, payload: HeartbeatA
    ) -> None:
        self.payload = payload
        if payload.TypeName == HeartbeatA_Maker.type_name:
            if from_role != GNodeRole.Supervisor:
                LOGGER.info(
                    f"Ignoring HeartbeatA from GNode with role {from_role}; expects Supervisor"
                )
            try:
                self.heartbeat_from_super(from_alias, payload)
            except:
                LOGGER.exception("Error in heartbeat_received")
        elif payload.TypeName == HeartbeatB_Maker.type_name:
            if from_role != GNodeRole.Scada:
                LOGGER.info(
                    f"Ignoring HeartbeatB from GNode with role {from_role}; expects Scada"
                )
            try:
                self.heartbeat_from_scada(payload)
            except:
                LOGGER.exception("Error in heartbeat_from_partner")
        elif payload.TypeName == JoinDispatchContract_Maker.type_name:
            if from_role != GNodeRole.Scada:
                LOGGER.info(
                    f"Ignoring JoinDispatchContract from GNode with role {from_role}; expects Scada"
                )
            try:
                self.join_dispatch_contract_from_scada(payload)
            except:
                LOGGER.exception("join_dispatch_contract_from_scada")
        elif payload.TypeName == LatestPrice_Maker.type_name:
            if from_role == GNodeRole.MarketMaker:
                try:
                    self.latest_price_from_market_maker(payload)
                except:
                    LOGGER.exception("Error in latest_price_from_market_maker")
        elif payload.TypeName == SimTimestep_Maker.type_name:
            try:
                self.timestep_from_timecoordinator(payload)
            except:
                LOGGER.exception("Error in timestep_from_timecoordinator")

    def heartbeat_from_super(self, from_alias: str, ping: HeartbeatA) -> None:
        pong = HeartbeatA_Maker(
            my_hex=str(random.choice("0123456789abcdef")), your_last_hex=ping.MyHex
        ).tuple

        self.send_message(
            payload=pong,
            to_role=GNodeRole.Supervisor,
            to_g_node_alias=self.settings.my_super_alias,
        )

        LOGGER.debug(
            f"[{self.alias}] Sent HB: SuHex {pong.YourLastHex}, AtnHex {pong.MyHex}"
        )

    def heartbeat_from_scada(self, ping: HeartbeatB) -> None:
        """
        This is the Atn's half of the DispatchContract Heartbeat pattern.
        It:
          - Checks that it has opted into a DispatchContract
          - Checks the  FromGNodeInstanceId to validate Scada credentials
          - Checks that the scada correctly repeated last sent hex
          - Updates the last hex received (for use at the top of the next minute) along w time received
        [more info](https://gridworks.readthedocs.io/en/latest/dispatch-contract.html)

        Args:
            payload (HeartbeatB): The latest heartbeat received from its
            SCADA partner

        """
        if not self.in_dispatch_contract():
            LOGGER.info(f"Not in Dispatch Contract. Ignoring")
        if ping.FromGNodeInstanceId != self.scada_gni_id:
            raise Exception("In dispatch contract but mismatched Scada Gni Id!")
        if ping.YourLastHex != self.hb_status.AtnLastHex:
            LOGGER.info(f"Received incorrect pong. Ignoring")
            return
        self.hb_status.LastHeartbeatReceivedMs = int(time.time() * 1000)
        self.hb_status.ScadaLastHex = ping.MyHex
        self.hb_status.AtnLastHex = str(random.choice("0123456789abcdef"))
        LOGGER.info(f"Got {ping.MyHex}")
        # Does not send back. Atn waits for the DispatchContract's expected
        # one minute before sending.
        # print(f"Got heartbeat from scada: {ping}")

    def hb_to_scada(self):
        """Checks that Atn is in Dispatch Contract, sends a HeartbeatB to Scada,
        and then reports to DispatchContract"""
        if not self.in_dispatch_contract():
            LOGGER.info("Not sending hb to Scada - not in DispatchContract yet")
        ping = HeartbeatB_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            my_hex=self.hb_status.AtnLastHex,
            your_last_hex=self.hb_status.ScadaLastHex,
            last_received_time_unix_ms=self.hb_status.LastHeartbeatReceivedMs,
            send_time_unix_ms=int(1000 * time.time()),
        ).tuple
        self.send_message(
            payload=ping,
            to_role=GNodeRole.Scada,
            to_g_node_alias=self.scada_alias,
        )
        LOGGER.info(f"Sent  {ping.MyHex}")
        # report to DispatchContract
        ptxn = PaymentTxn(self.acct.addr, self.sp, self.dc_client.app_addr, 1000)
        self.dc_client.call(
            DispatchContract.heartbeat_algo_audit,
            signed_proof=TransactionWithSigner(ptxn, self.acct.as_signer()),
            heartbeat=ping.as_dict(),
        )

    def check_for_dispatch_contract(self):
        """Upon startup, see if already party of a dispatch contract,
        and if so start up the dc_client and set the dc_app_id"""
        apps = self.client.account_info(self.acct.addr)["apps-local-state"]
        if len(apps) > 1:
            raise Exception("Should only be part of one app")
        if len(apps) == 0:
            return
        self.dc_app_id = apps[0]["id"]
        self.dc_client = ApplicationClient(
            client=self.client,
            app=DispatchContract(),
            signer=self.acct.as_signer(),
            app_id=self.dc_app_id,
        )

    def update_trading_rights(self) -> None:
        """Look at address to find our trading rights"""
        balances = algo_utils.get_balances(self.client, self.acct.addr)
        non_algo_asas = list(set(balances.keys()) - {0})
        ta_trading_rights = list(
            filter(
                lambda x: self.client.asset_info(x)["params"]["unit-name"] == "TATRADE",
                non_algo_asas,
            )
        )
        if len(ta_trading_rights) != 1:
            return
        x = ta_trading_rights[0]
        if self.client.asset_info(x)["params"]["name"] != self.ta_alias:
            raise Exception(
                f"Mismatch in trading rights and Alias. {self.alias} and {x}"
            )
        self.trading_rights_id = GwCertId_Maker(
            type=AlgoCertType.ASA, idx=ta_trading_rights[0], addr=None
        ).tuple

    def join_dispatch_contract_from_scada(self, payload: JoinDispatchContract) -> None:
        """Bootstraps the DispatchContract, opts in, and sends back DispatchContractConfirmed

        - ignores if already opted into a DispatchContract
        - Until World registry is in place, update g node instance id and write to .env
        - sets up ApplicationClient using the app_id sent by the scada
        - Checks that scada bootstrapping is done (contract has ta_alias and scada_cert_idx)
        - Check that our Atn acct owns the TaTradingRights for this TaAlias

        """
        if self.in_dispatch_contract():
            LOGGER.warning(
                f"Ignoring JoinDispatchContract - already in dispatch contract {self.dc_app_id}"
            )
            return
        self.scada_gni_id = payload.FromGNodeInstanceId
        with open(".env", "a") as file:
            file.write(f'ATN_SCADA_GNI_ID="{payload.FromGNodeInstanceId}"\n')
        self.dc_client = ApplicationClient(
            client=self.client,
            app=DispatchContract(),
            signer=self.acct.as_signer(),
            app_id=payload.DispatchContractAppId,
        )
        self.dc_app_id = payload.DispatchContractAppId
        app_state = self.dc_client.get_application_state()
        if not {"ta_alias", "scada_cert_idx"}.issubset(set(app_state.keys())):
            LOGGER.warning(
                f"AppId {payload.DispatchContractAppId} not done with scada bootstrap!"
            )
            return
        if not self.ta_alias == app_state["ta_alias"]:
            LOGGER.warning(
                f"Wrong TaAlias. Mine is {self.ta_alias}, DispatchContract has {app_state['ta_alias']}"
            )

        if self.trading_rights_id is None:
            self.update_trading_rights()
        if self.trading_rights_id is None:
            LOGGER.warning(f"Do not have trading rights! Cannot join dispatch contract")
            return
        ptxn = PaymentTxn(self.acct.addr, self.sp, self.dc_client.app_addr, 100_000)

        try:
            result = self.dc_client.call(
                DispatchContract.bootstrap2,
                atn_seed=TransactionWithSigner(ptxn, self.acct.as_signer()),
                TaTradingRights=self.trading_rights_id.Idx,
            )
        except Exception as e:
            LOGGER.warning(f"Failed to bootstrap DispatchContract: {e}")
            return

        # opts in
        self.dc_client.opt_in()
        LOGGER.info(f"Dispatch Contract {self.dc_app_id} is live")

        # sends back DispatchContractConfirmed

        txn = PaymentTxn(
            sender=self.acct.addr,
            receiver=self.dc_client.app_addr,
            sp=self.sp,
            amt=1000,
        )
        signed_txn = txn.sign(self.acct.sk)
        # don't need to actually send transaction for sig
        payload = DispatchContractConfirmed_Maker(
            from_g_node_alias=self.alias,
            from_g_node_instance_id=self.g_node_instance_id,
            signed_proof=encoding.msgpack_encode(signed_txn),
            atn_params=self.atn_params,
        ).tuple
        self.send_message(
            payload=payload, to_role=GNodeRole.Scada, to_g_node_alias=self.scada_alias
        )

    @abstractmethod
    def latest_price_from_market_maker(self, payload: LatestPrice) -> None:
        raise NotImplementedError

    def timestep_from_timecoordinator(self, payload: SimTimestep):
        if self._time == 0:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.info(f"TIME STARTED: {self.time_str()}")
        elif self._time < payload.TimeUnixS:
            self._time = payload.TimeUnixS
            self.new_timestep(payload)
            LOGGER.debug(f"Time is now {self.time_str()}")
        elif self._time == payload.TimeUnixS:
            self.repeat_timestep(payload)

    @abstractmethod
    def new_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("New timestep in atn_actor_base")
        raise NotImplementedError

    @abstractmethod
    def repeat_timestep(self, payload: SimTimestep) -> None:
        # LOGGER.info("Timestep received again in atn_actor_base")
        raise NotImplementedError

    def time_str(self) -> str:
        return pendulum.from_timestamp(self.time()).strftime("%m/%d/%Y, %H:%M")

    @property
    def ta_alias(self):
        """Add `ta` to the and of the Atn's GNodeAlias"""
        return self.alias + ".ta"

    @property
    def scada_alias(self):
        """Add `scada` from the end of the Atn's's GNodeAlias"""
        return self.alias + ".scada"

    def in_dispatch_contract(self) -> bool:
        """Checks that bootstrap 2 was completed w addition of ta_trading_rights_idx,
        and also that Scada is opted in"""
        if self.dc_client is None:
            return False
        app_state = self.dc_client.get_application_state()
        if "ta_trading_rights_idx" not in app_state.keys():
            return False
        # Todo: also check that atn is opted in
        apps = self.client.account_info(self.acct.addr)["apps-local-state"]
        app_ids = list(map(lambda x: x["id"], apps))
        if self.dc_client.app_id in app_ids:
            return True
        return False



import logging

import gridworks.algo_utils as algo_utils
import gridworks.api_utils as api_utils
import requests
from algosdk.v2client.algod import AlgodClient
from gridworks.algo_utils import BasicAccount
from gridworks.utils import RestfulResponse

import gwatn.config as config
from gwatn.enums import CoreGNodeRole
from gwatn.types import DiscoverycertAlgoCreate_Maker


LOGGER = logging.getLogger(__name__)


class DevDiscoverer:
    def __init__(self, settings: config.DiscovererSettings):
        self.settings = settings
        self.client: AlgodClient = AlgodClient(
            settings.algo_api_secrets.algod_token.get_secret_value(),
            settings.public.algod_address,
        )
        self.acct: BasicAccount = algo_utils.BasicAccount(
            private_key=self.settings.sk.get_secret_value()
        )
        self.multi: algo_utils.MultisigAccount = (
            api_utils.get_discoverer_account_with_admin(self.acct.addr)
        )
        # self.seed_fund_own_account()
        LOGGER.info("DevDiscoverer Initialized")

    def post_discoverycert_algo_create(self) -> RestfulResponse:
        payload = DiscoverycertAlgoCreate_Maker(
            g_node_alias=config.DiscovererSettings().discovered_ctn_alias,
            old_child_alias_list=config.DiscovererSettings().original_child_alias_list,
            discoverer_addr=self.acct.addr,
            supporting_material_hash="supporting material",
            core_g_node_role=CoreGNodeRole.ConductorTopologyNode,
            micro_lon=config.DiscovererSettings().micro_lon,
            micro_lat=config.DiscovererSettings().micro_lat,
        ).tuple
        api_endpoint = f"{config.GnfPublic().gnf_api_root}/discoverycert-algo-create/"
        r = requests.post(url=api_endpoint, json=payload.as_dict())



"""Type basegnode.scada.create, version 000"""
import json
from typing import Any
from typing import Dict
from typing import Literal

from gridworks.errors import SchemaError
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class BasegnodeScadaCreate(BaseModel):
    """Scada BaseGNode Creation.

    This is a payload designed to be sent from a TaOwner to the GNodeFactory. The TaOwner creates a private Algorand key and puts it on the Scada Device that will sense and control their TerminalAsset. The public address is associated with the Scada GNode by the GNodeFactory.
    """

    TaAlias: str = Field(
        title="TerminalAsset Alias",
        description="GNodeAlias of the TerminalAsset that will be controlled by the new SCADA GNode. The SCADA GNodeAlias will have '.scada' appended to this.",
    )
    ScadaAddr: str = Field(
        title="Algorand address for the SCADA",
        description="The TaOwner makes the corresponding private key, puts it on the SCADA device, and then sends this address to the GNodeFactory.",
    )
    TaDaemonAddr: str = Field(
        title="Algorand address of the associated TaDaemon",
        description="The TaDaemonAddr will have the TaDeed, and can be used to verify the public address of the TaOwner",
    )
    GNodeRegistryAddr: str = Field(
        title="GNodeRegistry Algorand address",
        description="The GNodeRegistry that contains Make/Model information about the SCADA and TerminalAsset",
    )
    SignedProof: str = Field(
        title="Recent transaction signed by the TaOwner",
        description="These will be replaced by composite transactions in next gen code.",
    )
    TypeName: Literal["basegnode.scada.create"] = "basegnode.scada.create"
    Version: str = "000"

    @validator("TaAlias")
    def _check_ta_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"TaAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("ScadaAddr")
    def _check_scada_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"ScadaAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("TaDaemonAddr")
    def _check_ta_daemon_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"TaDaemonAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("GNodeRegistryAddr")
    def _check_g_node_registry_addr(cls, v: str) -> str:
        try:
            check_is_algo_address_string_format(v)
        except ValueError as e:
            raise ValueError(
                f"GNodeRegistryAddr failed AlgoAddressStringFormat format validation: {e}"
            )
        return v

    @validator("SignedProof")
    def _check_signed_proof(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"SignedProof failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: TaOwner is SignedProof signer.
        The TaDaemonAddr provides the public address for the TaOwner. This TaOwnerAddr must match
        the signature on the SignedProof.
        """
        # TODO: Implement check for axiom 1"
        return v

    @root_validator
    def check_axiom_2(cls, v: dict) -> dict:
        """
        Axiom 2: TaAlias matches TaDeed.
        The TaDaemonAddr owns a TaDeed for the TaAlias.
        """
        # TODO: Implement check for axiom 2"
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class BasegnodeScadaCreate_Maker:
    type_name = "basegnode.scada.create"
    version = "000"

    def __init__(
        self,
        ta_alias: str,
        scada_addr: str,
        ta_daemon_addr: str,
        g_node_registry_addr: str,
        signed_proof: str,
    ):
        self.tuple = BasegnodeScadaCreate(
            TaAlias=ta_alias,
            ScadaAddr=scada_addr,
            TaDaemonAddr=ta_daemon_addr,
            GNodeRegistryAddr=g_node_registry_addr,
            SignedProof=signed_proof,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: BasegnodeScadaCreate) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> BasegnodeScadaCreate:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> BasegnodeScadaCreate:
        d2 = dict(d)
        if "TaAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TaAlias")
        if "ScadaAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing ScadaAddr")
        if "TaDaemonAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TaDaemonAddr")
        if "GNodeRegistryAddr" not in d2.keys():
            raise SchemaError(f"dict {d2} missing GNodeRegistryAddr")
        if "SignedProof" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SignedProof")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return BasegnodeScadaCreate(
            TaAlias=d2["TaAlias"],
            ScadaAddr=d2["ScadaAddr"],
            TaDaemonAddr=d2["TaDaemonAddr"],
            GNodeRegistryAddr=d2["GNodeRegistryAddr"],
            SignedProof=d2["SignedProof"],
            TypeName=d2["TypeName"],
            Version="000",
        )



"""Type atn.bid, version 001"""
import json
from enum import auto
from typing import Any
from typing import Dict
from typing import List
from typing import Literal

from fastapi_utils.enums import StrEnum
from gridworks.errors import SchemaError
from gridworks.message import as_enum
from pydantic import BaseModel
from pydantic import Field
from pydantic import root_validator
from pydantic import validator

from gwatn.enums import MarketPriceUnit
from gwatn.enums import MarketQuantityUnit
from gwatn.enums import MarketTypeName
from gwatn.types.price_quantity_unitless import PriceQuantityUnitless
from gwatn.types.price_quantity_unitless import PriceQuantityUnitless_Maker


class MarketPriceUnit000SchemaEnum:
    enum_name: str = "market.price.unit.000"
    symbols: List[str] = [
        "00000000",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketPriceUnit000(StrEnum):
    USDPerMWh = auto()

    @classmethod
    def default(cls) -> "MarketPriceUnit000":
        return cls.USDPerMWh

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketPriceUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketPriceUnit:
        if not MarketPriceUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketPriceUnit000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketPriceUnit, MarketPriceUnit.default())

    @classmethod
    def local_to_type(cls, market_price_unit: MarketPriceUnit) -> str:
        if not isinstance(market_price_unit, MarketPriceUnit):
            raise SchemaError(f"{market_price_unit} must be of type {MarketPriceUnit}")
        versioned_enum = as_enum(
            market_price_unit, MarketPriceUnit000, MarketPriceUnit000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketPriceUnit000] = {
        "00000000": MarketPriceUnit000.USDPerMWh,
    }

    versioned_enum_to_type_dict: Dict[MarketPriceUnit000, str] = {
        MarketPriceUnit000.USDPerMWh: "00000000",
    }


class MarketTypeName000SchemaEnum:
    enum_name: str = "market.type.name.000"
    symbols: List[str] = [
        "00000000",
        "d20b81e4",
        "b36cbfb4",
        "94a3fe9b",
        "5f335bdb",
        "01a84101",
        "e997ccfb",
        "618f9c0a",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketTypeName000(StrEnum):
    unknown = auto()
    rt5gate5 = auto()
    rt60gate5 = auto()
    da60 = auto()
    rt60gate30 = auto()
    rt15gate5 = auto()
    rt30gate5 = auto()
    rt60gate30b = auto()

    @classmethod
    def default(cls) -> "MarketTypeName000":
        return cls.unknown

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketTypeNameMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketTypeName:
        if not MarketTypeName000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketTypeName000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketTypeName, MarketTypeName.default())

    @classmethod
    def local_to_type(cls, market_type_name: MarketTypeName) -> str:
        if not isinstance(market_type_name, MarketTypeName):
            raise SchemaError(f"{market_type_name} must be of type {MarketTypeName}")
        versioned_enum = as_enum(
            market_type_name, MarketTypeName000, MarketTypeName000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketTypeName000] = {
        "00000000": MarketTypeName000.unknown,
        "d20b81e4": MarketTypeName000.rt5gate5,
        "b36cbfb4": MarketTypeName000.rt60gate5,
        "94a3fe9b": MarketTypeName000.da60,
        "5f335bdb": MarketTypeName000.rt60gate30,
        "01a84101": MarketTypeName000.rt15gate5,
        "e997ccfb": MarketTypeName000.rt30gate5,
        "618f9c0a": MarketTypeName000.rt60gate30b,
    }

    versioned_enum_to_type_dict: Dict[MarketTypeName000, str] = {
        MarketTypeName000.unknown: "00000000",
        MarketTypeName000.rt5gate5: "d20b81e4",
        MarketTypeName000.rt60gate5: "b36cbfb4",
        MarketTypeName000.da60: "94a3fe9b",
        MarketTypeName000.rt60gate30: "5f335bdb",
        MarketTypeName000.rt15gate5: "01a84101",
        MarketTypeName000.rt30gate5: "e997ccfb",
        MarketTypeName000.rt60gate30b: "618f9c0a",
    }


class MarketQuantityUnit000SchemaEnum:
    enum_name: str = "market.quantity.unit.000"
    symbols: List[str] = [
        "00000000",
        "c272f3b3",
    ]

    @classmethod
    def is_symbol(cls, candidate: str) -> bool:
        if candidate in cls.symbols:
            return True
        return False


class MarketQuantityUnit000(StrEnum):
    AvgMW = auto()
    AvgkW = auto()

    @classmethod
    def default(cls) -> "MarketQuantityUnit000":
        return cls.AvgMW

    @classmethod
    def values(cls) -> List[str]:
        return [elt.value for elt in cls]


class MarketQuantityUnitMap:
    @classmethod
    def type_to_local(cls, symbol: str) -> MarketQuantityUnit:
        if not MarketQuantityUnit000SchemaEnum.is_symbol(symbol):
            raise SchemaError(f"{symbol} must belong to MarketQuantityUnit000 symbols")
        versioned_enum = cls.type_to_versioned_enum_dict[symbol]
        return as_enum(versioned_enum, MarketQuantityUnit, MarketQuantityUnit.default())

    @classmethod
    def local_to_type(cls, market_quantity_unit: MarketQuantityUnit) -> str:
        if not isinstance(market_quantity_unit, MarketQuantityUnit):
            raise SchemaError(
                f"{market_quantity_unit} must be of type {MarketQuantityUnit}"
            )
        versioned_enum = as_enum(
            market_quantity_unit, MarketQuantityUnit000, MarketQuantityUnit000.default()
        )
        return cls.versioned_enum_to_type_dict[versioned_enum]

    type_to_versioned_enum_dict: Dict[str, MarketQuantityUnit000] = {
        "00000000": MarketQuantityUnit000.AvgMW,
        "c272f3b3": MarketQuantityUnit000.AvgkW,
    }

    versioned_enum_to_type_dict: Dict[MarketQuantityUnit000, str] = {
        MarketQuantityUnit000.AvgMW: "00000000",
        MarketQuantityUnit000.AvgkW: "c272f3b3",
    }


def check_is_uuid_canonical_textual(v: str) -> None:
    """
    UuidCanonicalTextual format:  A string of hex words separated by hyphens
    of length 8-4-4-4-12.

    Raises:
        ValueError: if not UuidCanonicalTextual format
    """
    try:
        x = v.split("-")
    except AttributeError as e:
        raise ValueError(f"Failed to split on -: {e}")
    if len(x) != 5:
        raise ValueError(f"{v} split by '-' did not have 5 words")
    for hex_word in x:
        try:
            int(hex_word, 16)
        except ValueError:
            raise ValueError(f"Words of {v} are not all hex")
    if len(x[0]) != 8:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[1]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[2]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[3]) != 4:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")
    if len(x[4]) != 12:
        raise ValueError(f"{v} word lengths not 8-4-4-4-12")


def check_is_left_right_dot(v: str) -> None:
    """
    LeftRightDot format: Lowercase alphanumeric words separated by periods,
    most significant word (on the left) starting with an alphabet character.

    Raises:
        ValueError: if not LeftRightDot format
    """
    from typing import List

    try:
        x: List[str] = v.split(".")
    except:
        raise ValueError(f"Failed to seperate {v} into words with split'.'")
    first_word = x[0]
    first_char = first_word[0]
    if not first_char.isalpha():
        raise ValueError(f"Most significant word of {v} must start with alphabet char.")
    for word in x:
        if not word.isalnum():
            raise ValueError(f"words of {v} split by by '.' must be alphanumeric.")
    if not v.islower():
        raise ValueError(f"All characters of {v} must be lowercase.")


def check_is_market_type_name_lrd_format(v: str) -> None:
    from gwatn.enums import MarketTypeName

    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    if not x[0] in MarketTypeName.values():
        raise ValueError(f"{v} not recognized MarketType")
    g_node_alias = ".".join(x[1:])
    check_is_left_right_dot(g_node_alias)


def check_is_market_slot_name_lrd_format(v: str) -> None:
    """
    MaketSlotNameLrdFormat: the format of a MarketSlotName.
      - The first word must be a MarketTypeName
      - The last word (unix time of market slot start) must
      be a 10-digit integer divisible by 300 (i.e. all MarketSlots
      start at the top of 5 minutes)
      - More strictly, the last word must be the start of a
      MarketSlot for that MarketType (i.e. divisible by 3600
      for hourly markets)
      - The middle words have LeftRightDot format (GNodeAlias
      of the MarketMaker)
    Example: rt60gate5.d1.isone.ver.keene.1673539200

    """
    from gwatn.data_classes import MarketType

    try:
        x = v.split(".")
    except AttributeError:
        raise ValueError(f"{v} failed to split on '.'")
    slot_start = x[-1]
    if len(slot_start) != 10:
        raise ValueError(f"slot start {slot_start} not of length 10")
    try:
        slot_start = int(slot_start)
    except ValueError:
        raise ValueError(f"slot start {slot_start} not an int")
    if slot_start % 300 != 0:
        raise ValueError(f"slot start {slot_start} not a multiple of 300")

    market_type_name_lrd = ".".join(x[:-1])
    try:
        check_is_market_type_name_lrd_format(market_type_name_lrd)
    except ValueError as e:
        raise ValueError(f"e")

    market_type = MarketType.by_id[market_type_name_lrd.split(".")[0]]
    if not slot_start % (market_type.duration_minutes * 60) == 0:
        raise ValueError(
            f"market_slot_start_s mod {(market_type.duration_minutes * 60)} must be 0"
        )


def check_is_algo_address_string_format(v: str) -> None:
    """
    AlgoAddressStringFormat format: The public key of a private/public Ed25519
    key pair, transformed into an  Algorand address, by adding a 4-byte checksum
    to the end of the public key and then encoding in base32.

    Raises:
        ValueError: if not AlgoAddressStringFormat format
    """
    import algosdk

    at = algosdk.abi.AddressType()
    try:
        result = at.decode(at.encode(v))
    except Exception as e:
        raise ValueError(f"Not AlgoAddressStringFormat: {e}")


def check_is_algo_msg_pack_encoded(v: str) -> None:
    """
    AlgoMSgPackEncoded format: the format of an  transaction sent to
    the Algorand blockchain.

    Raises:
        ValueError: if not AlgoMSgPackEncoded  format
    """
    import algosdk

    try:
        algosdk.encoding.future_msgpack_decode(v)
    except Exception as e:
        raise ValueError(f"Not AlgoMsgPackEncoded format: {e}")


class AtnBid(BaseModel):
    """AtomicTNode bid sent to a MarketMaker
    [More info](https://gridworks.readthedocs.io/en/latest/market-bid.html).
    """

    BidderAlias: str = Field(
        title="BidderAlias",
    )
    BidderGNodeInstanceId: str = Field(
        title="BidderGNodeInstanceId",
    )
    MarketSlotName: str = Field(
        title="MarketSlotName",
    )
    PqPairs: List[PriceQuantityUnitless] = Field(
        title="Price Quantity Pairs",
        description="The list of Price Quantity Pairs making up the bid. The units are provided by the AtnBid.PriceUnit and AtnBid.QuantityUnit.",
    )
    InjectionIsPositive: bool = Field(
        title="InjectionIsPositive",
        default=False,
    )
    PriceUnit: MarketPriceUnit = Field(
        title="PriceUnit",
        default=MarketPriceUnit.USDPerMWh,
    )
    QuantityUnit: MarketQuantityUnit = Field(
        title="QuantityUnit",
        default=MarketQuantityUnit.AvgMW,
    )
    SignedMarketFeeTxn: str = Field(
        title="SignedMarketFeeTxn",
    )
    TypeName: Literal["atn.bid"] = "atn.bid"
    Version: str = "001"

    @validator("BidderAlias")
    def _check_bidder_alias(cls, v: str) -> str:
        try:
            check_is_left_right_dot(v)
        except ValueError as e:
            raise ValueError(f"BidderAlias failed LeftRightDot format validation: {e}")
        return v

    @validator("BidderGNodeInstanceId")
    def _check_bidder_g_node_instance_id(cls, v: str) -> str:
        try:
            check_is_uuid_canonical_textual(v)
        except ValueError as e:
            raise ValueError(
                f"BidderGNodeInstanceId failed UuidCanonicalTextual format validation: {e}"
            )
        return v

    @validator("MarketSlotName")
    def _check_market_slot_name(cls, v: str) -> str:
        try:
            check_is_market_slot_name_lrd_format(v)
        except ValueError as e:
            raise ValueError(
                f"MarketSlotName failed MarketSlotNameLrdFormat format validation: {e}"
            )
        return v

    @validator("PqPairs")
    def _check_pq_pairs(cls, v: List) -> List:
        for elt in v:
            if not isinstance(elt, PriceQuantityUnitless):
                raise ValueError(
                    f"elt {elt} of PqPairs must have type PriceQuantityUnitless."
                )
        return v

    @validator("PriceUnit")
    def _check_price_unit(cls, v: MarketPriceUnit) -> MarketPriceUnit:
        return as_enum(v, MarketPriceUnit, MarketPriceUnit.USDPerMWh)

    @validator("QuantityUnit")
    def _check_quantity_unit(cls, v: MarketQuantityUnit) -> MarketQuantityUnit:
        return as_enum(v, MarketQuantityUnit, MarketQuantityUnit.AvgMW)

    @validator("SignedMarketFeeTxn")
    def _check_signed_market_fee_txn(cls, v: str) -> str:
        try:
            check_is_algo_msg_pack_encoded(v)
        except ValueError as e:
            raise ValueError(
                f"SignedMarketFeeTxn failed AlgoMsgPackEncoded format validation: {e}"
            )
        return v

    @root_validator
    def check_axiom_1(cls, v: dict) -> dict:
        """
        Axiom 1: PqPairs PriceMax matches MarketType.
        There is a GridWorks global list of MarketTypes (a GridWorks type), identified by their MarketTypeNames (a GridWorks enum).  The MarketType has a PriceMax, which must be the first price of the first PriceQuantity pair in PqPairs.
        """
        # TODO: Implement check for axiom 1"
        return v

    @root_validator
    def check_axiom_2(cls, v: dict) -> dict:
        """
        Axiom 2: .

        """
        # TODO: Implement check for axiom 2"
        return v

    def as_dict(self) -> Dict[str, Any]:
        d = self.dict()

        # Recursively call as_dict() for the SubTypes
        pq_pairs = []
        for elt in self.PqPairs:
            pq_pairs.append(elt.as_dict())
        d["PqPairs"] = pq_pairs
        del d["PriceUnit"]
        PriceUnit = as_enum(self.PriceUnit, MarketPriceUnit, MarketPriceUnit.default())
        d["PriceUnitGtEnumSymbol"] = MarketPriceUnitMap.local_to_type(PriceUnit)
        del d["QuantityUnit"]
        QuantityUnit = as_enum(
            self.QuantityUnit, MarketQuantityUnit, MarketQuantityUnit.default()
        )
        d["QuantityUnitGtEnumSymbol"] = MarketQuantityUnitMap.local_to_type(
            QuantityUnit
        )
        return d

    def as_type(self) -> str:
        return json.dumps(self.as_dict())

    def __hash__(self):
        return hash((type(self),) + tuple(self.__dict__.values()))  # noqa


class AtnBid_Maker:
    type_name = "atn.bid"
    version = "001"

    def __init__(
        self,
        bidder_alias: str,
        bidder_g_node_instance_id: str,
        market_slot_name: str,
        pq_pairs: List[PriceQuantityUnitless],
        injection_is_positive: bool,
        price_unit: MarketPriceUnit,
        quantity_unit: MarketQuantityUnit,
        signed_market_fee_txn: str,
    ):
        self.tuple = AtnBid(
            BidderAlias=bidder_alias,
            BidderGNodeInstanceId=bidder_g_node_instance_id,
            MarketSlotName=market_slot_name,
            PqPairs=pq_pairs,
            InjectionIsPositive=injection_is_positive,
            PriceUnit=price_unit,
            QuantityUnit=quantity_unit,
            SignedMarketFeeTxn=signed_market_fee_txn,
            #
        )

    @classmethod
    def tuple_to_type(cls, tuple: AtnBid) -> str:
        """
        Given a Python class object, returns the serialized JSON type object
        """
        return tuple.as_type()

    @classmethod
    def type_to_tuple(cls, t: str) -> AtnBid:
        """
        Given a serialized JSON type object, returns the Python class object
        """
        try:
            d = json.loads(t)
        except TypeError:
            raise SchemaError("Type must be string or bytes!")
        if not isinstance(d, dict):
            raise SchemaError(f"Deserializing {t} must result in dict!")
        return cls.dict_to_tuple(d)

    @classmethod
    def dict_to_tuple(cls, d: dict[str, Any]) -> AtnBid:
        d2 = dict(d)
        if "BidderAlias" not in d2.keys():
            raise SchemaError(f"dict {d2} missing BidderAlias")
        if "BidderGNodeInstanceId" not in d2.keys():
            raise SchemaError(f"dict {d2} missing BidderGNodeInstanceId")
        if "MarketSlotName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing MarketSlotName")
        if "PqPairs" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PqPairs")
        pq_pairs = []
        if not isinstance(d2["PqPairs"], List):
            raise SchemaError("PqPairs must be a List!")
        for elt in d2["PqPairs"]:
            if not isinstance(elt, dict):
                raise SchemaError(
                    f"elt {elt} of PqPairs must be "
                    "PriceQuantityUnitless but not even a dict!"
                )
            pq_pairs.append(PriceQuantityUnitless_Maker.dict_to_tuple(elt))
        d2["PqPairs"] = pq_pairs
        if "InjectionIsPositive" not in d2.keys():
            raise SchemaError(f"dict {d2} missing InjectionIsPositive")
        if "PriceUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing PriceUnitGtEnumSymbol")
        if d2["PriceUnitGtEnumSymbol"] in MarketPriceUnit000SchemaEnum.symbols:
            d2["PriceUnit"] = MarketPriceUnitMap.type_to_local(
                d2["PriceUnitGtEnumSymbol"]
            )
        else:
            d2["PriceUnit"] = MarketPriceUnit.default()
        if "QuantityUnitGtEnumSymbol" not in d2.keys():
            raise SchemaError(f"dict {d2} missing QuantityUnitGtEnumSymbol")
        if d2["QuantityUnitGtEnumSymbol"] in MarketQuantityUnit000SchemaEnum.symbols:
            d2["QuantityUnit"] = MarketQuantityUnitMap.type_to_local(
                d2["QuantityUnitGtEnumSymbol"]
            )
        else:
            d2["QuantityUnit"] = MarketQuantityUnit.default()
        if "SignedMarketFeeTxn" not in d2.keys():
            raise SchemaError(f"dict {d2} missing SignedMarketFeeTxn")
        if "TypeName" not in d2.keys():
            raise SchemaError(f"dict {d2} missing TypeName")

        return AtnBid(
            BidderAlias=d2["BidderAlias"],
            BidderGNodeInstanceId=d2["BidderGNodeInstanceId"],
            MarketSlotName=d2["MarketSlotName"],
            PqPairs=d2["PqPairs"],
            InjectionIsPositive=d2["InjectionIsPositive"],
            PriceUnit=d2["PriceUnit"],
            QuantityUnit=d2["QuantityUnit"],
            SignedMarketFeeTxn=d2["SignedMarketFeeTxn"],
            TypeName=d2["TypeName"],
            Version="001",
        )



"""The Rest API wrapper for the TaDaemon Smart Contract"""
from fastapi import FastAPI
from fastapi import HTTPException
from gridworks.utils import RestfulResponse

from gwatn.python_ta_daemon import PythonTaDaemon
from gwatn.types import InitialTadeedAlgoOptin
from gwatn.types import NewTadeedAlgoOptin
from gwatn.types import OldTadeedAlgoReturn
from gwatn.types import SlaEnter


tags_metadata = [
    {
        "name": "initial.tadeed.algo.optin",
        "description": "Request TaDaemon to opt into the first TaDeed for its TerminalAsset. "
        "Must be sent by the TaOwner.",
        "externalDocs": {
            "description": "Lexicon w definitions of TaDaemon, TaDeed, TerminalAsset, TaOwner",
            "url": "https://gridworks.readthedocs.io/en/latest/lexicon.html",
        },
    },
]

app = FastAPI(openapi_tags=tags_metadata)
daemon = PythonTaDaemon()

# @app.get("/", response_class=FileResponse)
# async def root():
#     if os.path.isfile("docs/wiki/img/mushroom.png"):
#         return FileResponse("docs/wiki/img/mushroom.png", media_type="image/png")
#     else:
#         return FileResponse("docs/wiki/img/mushroom.png", media_type="image/png")


# @app.get("/icon/")
# async def main():
#     return daemon.ta_deed_icon_file


@app.get("/owned-tadeeds/")
async def main():
    return {"Owned deeds": daemon.ta_deed_alias_list()}


@app.get("/trading-rights/")
async def main():
    if daemon.trading_rights_addr == daemon.acct.addr:
        return {"Address of Trading Rights Owner": "self"}
    else:
        return {"Address of Trading Rights Owner": daemon.trading_rights_addr}


@app.get("/env/")
async def show_env():
    return daemon.settings


@app.post(
    "/initial-tadeed-algo-optin/",
    response_model=RestfulResponse,
    tags=["initial.tadeed.algo.optin"],
)
async def initial_tadeed_algo_optin_received(payload: InitialTadeedAlgoOptin):
    r = daemon.initial_tadeed_algo_optin_received(payload)
    if r.HttpStatusCode > 200:
        raise HTTPException(
            status_code=r.HttpStatusCode, detail=f"[{r.HttpStatusCode}]: {r.Note}"
        )
    return r


@app.post("/new-tadeed-algo-optin/", response_model=RestfulResponse)
async def new_tadeed_algo_received(payload: NewTadeedAlgoOptin):
    r = daemon.new_tadeed_algo_optin_received(payload)
    if r.HttpStatusCode > 200:
        raise HTTPException(
            status_code=r.HttpStatusCode, detail=f"[{r.HttpStatusCode}]: {r.Note}"
        )
    return r


@app.post("/old-tadeed-algo-return/", response_model=RestfulResponse)
async def old_tadeed_algo_return_received(payload: OldTadeedAlgoReturn):
    r = daemon.old_tadeed_algo_return_received(payload)
    if r.HttpStatusCode > 200:
        raise HTTPException(
            status_code=r.HttpStatusCode, detail=f"[{r.HttpStatusCode}]: {r.Note}"
        )
    return r


@app.post("/sla-enter/", response_model=RestfulResponse)
async def old_tadeed_algo_return_received(payload: SlaEnter):
    r = daemon.sla_enter_received(payload)
    if r.HttpStatusCode > 200:
        raise HTTPException(
            status_code=r.HttpStatusCode, detail=f"[{r.HttpStatusCode}]: {r.Note}"
        )
    return r
